from typing import Any, Union, Callable, Iterable, Optional
from typing_extensions import Literal

from copy import copy
from enum import auto
from math import fsum

from anndata import AnnData
from cellrank import logging as logg
from cellrank.tl._enum import _DEFAULT_BACKEND, ModeEnum
from cellrank.ul._docs import d, inject_docs
from cellrank.ul._utils import valuedispatch
from cellrank.tl.kernels import Kernel
from cellrank.ul._parallelize import parallelize
from cellrank.tl.kernels._utils import (
    prange,
    np_mean,
    _filter_kwargs,
    _random_normal,
    _reconstruct_one,
    _calculate_starts,
    _get_probs_for_zero_vec,
)
from scvelo.preprocessing.moments import get_moments
from cellrank.tl.kernels._base_kernel import _RTOL
from cellrank.tl.kernels._velocity_schemes import Scheme, _get_scheme

import numpy as np
from scipy.sparse import issparse, csr_matrix


class VelocityMode(ModeEnum):  # noqa: D101
    DETERMINISTIC = auto()
    STOCHASTIC = auto()
    SAMPLING = auto()
    MONTE_CARLO = auto()


class BackwardMode(ModeEnum):  # noqa: D101
    TRANSPOSE = auto()
    NEGATE = auto()


@d.dedent
class VelocityKernel(Kernel):
    """
    Kernel which computes a transition matrix based on RNA velocity.

    This borrows ideas from both :cite:`manno:18` and :cite:`bergen:20`. In short, for each cell *i*, we compute
    transition probabilities :math:`p_{i, j}` to each cell *j* in the neighborhood of *i*. The transition probabilities
    are computed as a multinomial logistic regression where the weights :math:`w_j` (for all *j*) are given
    by the vector that connects cell *i* with cell *j* in gene expression space, and the features :math:`x_i` are given
    by the velocity vector :math:`v_i` of cell *i*.

    Parameters
    ----------
    %(adata)s
    %(backward)s
    vkey
        Key in :attr:`anndata.AnnData.layers` where velocities are stored.
    xkey
        Key in :attr:`anndata.AnnData.layers` where expected gene expression counts are stored.
    gene_subset
        List of genes to be used to compute transition probabilities.
        By default, genes from :attr:`anndata.AnnData.var` ``['velocity_genes']`` are used.
    %(cond_num)s
    check_connectivity
        Check whether the underlying KNN graph is connected.
    kwargs
        Keyword arguments for :class:`cellrank.tl.kernels.Kernel`.
    """

    def __init__(
        self,
        adata: AnnData,
        backward: bool = False,
        vkey: str = "velocity",
        xkey: str = "Ms",
        gene_subset: Optional[Iterable] = None,
        compute_cond_num: bool = False,
        check_connectivity: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            adata,
            backward=backward,
            vkey=vkey,
            xkey=xkey,
            gene_subset=gene_subset,
            compute_cond_num=compute_cond_num,
            check_connectivity=check_connectivity,
            **kwargs,
        )
        self._vkey = vkey  # for copy
        self._xkey = xkey
        self._gene_subset = gene_subset
        self._logits = None

    def _read_from_adata(self, **kwargs: Any) -> None:
        super()._read_from_adata(**kwargs)

        # check whether velocities have been computed
        vkey = kwargs.pop("vkey", "velocity")
        if vkey not in self.adata.layers.keys():
            raise KeyError("Compute RNA velocity first as `scvelo.tl.velocity()`.")

        # restrict genes to a subset, i.e. velocity genes or user provided list
        gene_subset = kwargs.pop("gene_subset", None)
        subset = np.ones(self.adata.n_vars, bool)
        if gene_subset is not None:
            var_names_subset = self.adata.var_names.isin(gene_subset)
            subset &= var_names_subset if len(var_names_subset) > 0 else gene_subset
        elif f"{vkey}_genes" in self.adata.var.keys():
            subset &= np.array(self.adata.var[f"{vkey}_genes"].values, dtype=bool)

        # choose data representation to use for transcriptomic displacements
        xkey = kwargs.pop("xkey", "Ms")
        xkey = xkey if xkey in self.adata.layers.keys() else "spliced"

        # filter both the velocities and the gene expression profiles to the gene subset and densify the matrices
        X = np.array(
            self.adata.layers[xkey].A[:, subset]
            if issparse(self.adata.layers[xkey])
            else self.adata.layers[xkey][:, subset]
        )
        V = np.array(
            self.adata.layers[vkey].A[:, subset]
            if issparse(self.adata.layers[vkey])
            else self.adata.layers[vkey][:, subset]
        )

        # remove genes that have any Nan values (in both X and V)
        nans = np.isnan(np.sum(V, axis=0))
        if np.any(nans):
            X = X[:, ~nans]
            V = V[:, ~nans]

        # check the velocity parameters
        par_key = f"{vkey}_params"
        velocity_params = self.adata.uns.get(par_key, None)
        if velocity_params is None:
            logg.debug(
                f"Unable to load velocity parameters from `adata.uns[{par_key!r}]`"
            )

        # add to self
        self._velocity = V.astype(np.float64)
        self._gene_expression = X.astype(np.float64)
        self._velocity_params = velocity_params

    @inject_docs(m=VelocityMode, b=BackwardMode, s=Scheme)  # don't swap the order
    @d.dedent
    def compute_transition_matrix(
        self,
        mode: Literal[
            "deterministic", "stochastic", "sampling", "monte_carlo"
        ] = VelocityMode.DETERMINISTIC,
        backward_mode: Literal["transpose", "negate"] = BackwardMode.TRANSPOSE,
        scheme: Union[
            Literal["dot_product", "cosine", "correlation"], Callable
        ] = Scheme.CORRELATION,
        softmax_scale: Optional[float] = None,
        n_samples: int = 1000,
        seed: Optional[int] = None,
        check_irreducibility: bool = False,
        **kwargs: Any,
    ) -> "VelocityKernel":
        """
        Compute transition matrix based on velocity directions on the local manifold.

        For each cell, infer transition probabilities based on the cell's velocity-extrapolated cell state and the
        cell states of its *K* nearest neighbors.

        Parameters
        ----------
        %(velocity_mode)s
        %(velocity_backward_mode)s
        %(softmax_scale)s
        %(velocity_scheme)s
        n_samples
            Number of bootstrap samples when ``mode = {m.MONTE_CARLO!r}``.
        seed
            Set the seed for random state when the method requires ``n_samples``.
        check_irreducibility
            Optional check for irreducibility of the final transition matrix.
        %(parallel)s

        Returns
        -------
        Self and updates the following fields:

            - :attr:`transition_matrix`.
            - :attr:`logits`.
        """
        mode = VelocityMode(mode)
        backward_mode = BackwardMode(backward_mode)

        if isinstance(scheme, str):
            scheme = _get_scheme(Scheme(scheme))

        if not callable(scheme):
            raise TypeError(
                f"Expected `scheme` to be a function, found `{type(scheme)!r}`."
            )

        if self.backward and mode != VelocityMode.DETERMINISTIC:
            logg.warning(
                f"Mode `{mode!r}` is currently not supported for the backward process. "
                f"Defaulting to mode `{VelocityMode.DETERMINISTIC!r}`"
            )
            mode = VelocityMode.DETERMINISTIC

        if mode == VelocityMode.STOCHASTIC:
            try:
                if not hasattr(
                    scheme, "hessian"
                ):  # this can also raise in our velocity scheme definition
                    raise NotImplementedError
            except NotImplementedError:
                logg.warning(
                    f"Unable to detect a method for Hessian computation. If using predefined functions, consider "
                    f"installing `jax` as `pip install jax jaxlib`\n"
                    f"Defaulting to `mode={VelocityMode.MONTE_CARLO!r}` and `n_samples={n_samples}`"
                )
                mode = VelocityMode.MONTE_CARLO

        start = logg.info(
            f"Computing transition matrix based on logits using `{mode!r}` mode"
        )

        if seed is None:
            seed = np.random.randint(0, 2 ** 16)

        # fmt: off
        params = {"softmax_scale": softmax_scale, "mode": mode, "seed": seed, "scheme": str(scheme)}
        if self.backward:
            params["bwd_mode"] = backward_mode
        if self._reuse_cache(params, time=start):
            return self

        # compute first and second order moments to model the distribution of the velocity vector
        np.random.seed(seed)
        velocity_expectation = get_moments(self.adata, self._velocity, second_order=False).astype(np.float64)
        velocity_variance = get_moments(self.adata, self._velocity, second_order=True).astype(np.float64)
        # fmt: on

        if mode == VelocityMode.MONTE_CARLO and n_samples == 1:
            logg.debug("Setting mode to sampling because `n_samples=1`")
            mode = VelocityMode.SAMPLING

        backend = kwargs.pop("backend", _DEFAULT_BACKEND)
        if mode != VelocityMode.STOCHASTIC and backend == "multiprocessing":
            # this is because on jitting and pickling (cloudpickle, used by loky, handles it correctly)
            logg.warning(
                f"Multiprocessing backend is supported only for mode `{VelocityMode.STOCHASTIC!r}`. "
                f"Defaulting to `{_DEFAULT_BACKEND!r}`"
            )
            backend = _DEFAULT_BACKEND

        if softmax_scale is None:
            logg.info(
                f"Estimating `softmax_scale` using `{VelocityMode.DETERMINISTIC!r}` mode"
            )
            _, cmat = _dispatch_computation(
                VelocityMode.DETERMINISTIC,
                scheme=scheme,
                conn=self._conn,
                expression=self._gene_expression,
                velocity=self._velocity,
                expectation=velocity_expectation,
                variance=velocity_variance,
                softmax_scale=1.0,
                backward=self.backward,
                backward_mode=backward_mode,
                n_samples=n_samples,
                seed=seed,
                backend=backend,
                **kwargs,
            )
            softmax_scale = 1.0 / np.median(np.abs(cmat.data))
            params["softmax_scale"] = softmax_scale
            logg.info(f"Setting `softmax_scale={softmax_scale:.4f}`")

        tmat, cmat = _dispatch_computation(
            mode,
            scheme=scheme,
            conn=self._conn,
            expression=self._gene_expression,
            velocity=self._velocity,
            expectation=velocity_expectation,
            variance=velocity_variance,
            softmax_scale=softmax_scale,
            backward=self.backward,
            backward_mode=backward_mode,
            n_samples=n_samples,
            seed=seed,
            backend=backend,
            **kwargs,
        )
        self._compute_transition_matrix(
            tmat, density_normalize=False, check_irreducibility=check_irreducibility
        )
        self._logits = cmat

        logg.info("    Finish", time=start)

        return self

    @property
    def logits(self) -> csr_matrix:
        """Array of shape ``(n_cells, n_cells)`` containing the logits."""
        return self._logits

    @d.dedent
    def copy(self) -> "VelocityKernel":
        """%(copy)s"""  # noqa
        vk = VelocityKernel(
            self.adata,
            backward=self.backward,
            vkey=self._vkey,
            xkey=self._xkey,
            gene_subset=self._gene_subset,
        )
        vk._params = copy(self.params)
        vk._cond_num = self.condition_number
        vk._transition_matrix = copy(self._transition_matrix)
        vk._logits = copy(self.logits)

        return vk


@valuedispatch
def _dispatch_computation(mode, *_args, **_kwargs):
    raise NotImplementedError(mode)


def _run_in_parallel(fn: Callable, conn: csr_matrix, **kwargs) -> Any:
    fname = fn.__name__
    if fname == "_run_stochastic":
        ixs = np.argsort(np.array((conn != 0).sum(1)).ravel())[::-1]
    else:
        ixs = np.arange(conn.shape[0])
        np.random.shuffle(ixs)

    unit = (
        "sample" if (fname == "_run_mc") and kwargs.get("n_samples", 1) > 1 else "cell"
    )
    kwargs["indices"] = conn.indices
    kwargs["indptr"] = conn.indptr

    return parallelize(
        fn,
        ixs,
        as_array=False,
        extractor=lambda res: _reconstruct_one(np.concatenate(res, axis=-1), conn, ixs),
        unit=unit,
        **_filter_kwargs(parallelize, **kwargs),
    )(**_filter_kwargs(fn, **kwargs))


def _run_deterministic(
    ixs: Union[prange, np.ndarray],
    scheme: Callable,
    indices: np.ndarray,
    indptr: np.ndarray,
    expression: np.ndarray,
    velocity: np.ndarray,
    backward: bool = False,
    backward_mode: BackwardMode = BackwardMode.TRANSPOSE,
    softmax_scale: float = 1,
    queue=None,
) -> np.ndarray:

    starts = _calculate_starts(indptr, ixs)
    probs_cors = np.empty((2, starts[-1]))

    for i in prange(len(ixs)):
        ix = ixs[i]
        start, end = indptr[ix], indptr[ix + 1]

        nbhs_ixs = indices[start:end]
        n_neigh = len(nbhs_ixs)
        assert n_neigh, f"Cell {ix} does not have any neighbors."

        W = expression[nbhs_ixs, :] - expression[ix, :]

        if not backward or (backward and backward_mode == BackwardMode.NEGATE):
            v = np.expand_dims(velocity[ix], 0)

            # for the negate backward mode
            if backward:
                v *= -1

            if np.all(v == 0):
                p, c = _get_probs_for_zero_vec(len(nbhs_ixs))
            else:
                p, c = scheme(v, W, softmax_scale)
        else:
            # compute how likely all neighbors are to transition to this cell
            V = velocity[nbhs_ixs, :]
            p, c = scheme(V, -1 * W, softmax_scale)

        probs_cors[0, starts[i] : starts[i] + n_neigh] = p
        probs_cors[1, starts[i] : starts[i] + n_neigh] = c

        if queue is not None:
            queue.put(1)

    if queue is not None:
        queue.put(None)

    return probs_cors


def _run_mc(
    ixs: Optional[np.ndarray],
    scheme: Callable,
    indices: np.ndarray,
    indptr: np.ndarray,
    expression: np.ndarray,
    expectation: np.ndarray,
    variance: np.ndarray,
    n_samples: int = 1,
    softmax_scale: float = 1,
    seed: int = 0,
    queue=None,
) -> np.ndarray:

    starts = _calculate_starts(indptr, ixs)
    probs_cors = np.empty((2, starts[-1]))

    for i in prange(len(ixs)):
        ix = ixs[i]
        start, end = indptr[ix], indptr[ix + 1]

        np.random.seed(seed + ix)

        nbhs_ixs = indices[start:end]
        n_neigh = len(nbhs_ixs)
        assert n_neigh, f"Cell {ix} does not have any neighbors."

        # get the displacement matrix. Changing dimensions b/c varying numbers of neighbors slow down autograd
        W = expression[nbhs_ixs, :] - expression[ix, :]

        # much faster (1.8x than sampling only 1 if average)
        samples = _random_normal(
            expectation[ix],
            variance[ix],
            n_samples=n_samples,
        )

        probs_cors_tmp = np.empty((2, n_samples, n_neigh))
        for j in prange(n_samples):
            prob, cor = scheme(np.atleast_2d(samples[j]), W, softmax_scale)
            probs_cors_tmp[0, j, :] = prob
            probs_cors_tmp[1, j, :] = cor

        probs_cors[0, starts[i] : starts[i] + n_neigh] = np_mean(
            probs_cors_tmp[0], axis=0
        )
        probs_cors[1, starts[i] : starts[i] + n_neigh] = np_mean(
            probs_cors_tmp[1], axis=0
        )

        if queue is not None:
            queue.put(1)

    if queue is not None:
        queue.put(None)

    return probs_cors


def _run_stochastic(
    ixs: np.ndarray,
    scheme: Callable,
    indices: np.ndarray,
    indptr: np.ndarray,
    expression: np.ndarray,
    expectation: np.ndarray,
    variance: np.ndarray,
    softmax_scale: float = 1,
    queue=None,
) -> np.ndarray:
    if not hasattr(scheme, "hessian"):
        raise AttributeError("Velocity scheme doesn't have `hessian` attribute.")

    starts = _calculate_starts(indptr, ixs)
    probs_cors = np.empty((2, starts[-1]))
    n_genes = expression.shape[1]

    for i, ix in enumerate(ixs):
        start, end = indptr[ix], indptr[ix + 1]

        nbhs_ixs = indices[start:end]
        n_neigh = len(nbhs_ixs)
        assert n_neigh, f"Cell {ix} does not have any neighbors."

        v = expectation[ix]

        if np.all(v == 0):
            p, c = _get_probs_for_zero_vec(len(nbhs_ixs))
        else:
            # compute the Hessian tensor, and turn it into a matrix that has the diagonal elements in its rows
            W = expression[nbhs_ixs, :] - expression[ix, :]

            H = scheme.hessian(v, W, softmax_scale)
            if H.shape == (n_neigh, n_genes):
                H_diag = H
            elif H.shape == (n_neigh, n_genes, n_genes):
                H_diag = np.array([np.diag(h) for h in H])
            else:
                raise ValueError(
                    f"Expected full Hessian matrix of shape `{(n_neigh, n_genes, n_genes)}` "
                    f"or its diagonal of shape `{(n_neigh, n_genes)}`, found `{H.shape}`."
                )

            # compute zero order term
            p_0, c = scheme(v[None, :], W, softmax_scale)

            # compute second order term (note that the first order term cancels)
            p_2 = 0.5 * H_diag.dot(variance[ix])

            # combine both to give the second order Taylor approximation. Can sometimes be negative because we
            # neglected higher order terms, so force it to be non-negative
            p = np.clip(p_0 + p_2, a_min=0, a_max=1)

            mask = np.isnan(p)
            p[mask] = 0
            c[mask] = 0

            if np.all(p == 0):
                p, c = _get_probs_for_zero_vec(len(nbhs_ixs))

            sum_ = fsum(p)
            if not np.isclose(sum_, 1.0, rtol=_RTOL):
                p[~mask] = p[~mask] / sum_

        probs_cors[0, starts[i] : starts[i] + n_neigh] = p
        probs_cors[1, starts[i] : starts[i] + n_neigh] = c

        if queue is not None:
            queue.put(1)

    if queue is not None:
        queue.put(None)

    return probs_cors


@_dispatch_computation.register(VelocityMode.SAMPLING)
def _(**kwargs):
    # cant 't use partial - parallelize needs to know whether function has `py_func` attribute + doesn't have a __name__
    kwargs["n_samples"] = 1
    return _run_in_parallel(_run_mc, **kwargs)


_dispatch_computation.register(VelocityMode.STOCHASTIC)(
    lambda **kwargs: _run_in_parallel(_run_stochastic, **kwargs)
)
_dispatch_computation.register(VelocityMode.DETERMINISTIC)(
    lambda **kwargs: _run_in_parallel(_run_deterministic, **kwargs)
)
_dispatch_computation.register(VelocityMode.MONTE_CARLO)(
    lambda **kwargs: _run_in_parallel(_run_mc, **kwargs)
)



"""
Plot aggregated cellular fates
------------------------------

This example shows how to aggregate fate probabilities from the single cell level to a cluster level
and how to visualize these in various ways.
"""

import scanpy as sc
import cellrank as cr

adata = cr.datasets.pancreas_preprocessed("../example.h5ad")
adata

# %%
# First, we compute the terminal states and the lineages, as well as :mod:`scanpy`'s PAGA :cite:`wolf:19`.
cr.tl.terminal_states(
    adata,
    cluster_key="clusters",
    weight_connectivities=0.2,
    n_states=3,
    softmax_scale=4,
    show_progress_bar=False,
)
cr.tl.lineages(adata)

sc.tl.paga(adata, "clusters")

# %%
# We can visualize the aggregate absorption probabilities as a bar plot. The whiskers correspond to the standard error
# of the mean.
cr.pl.cluster_fates(adata, mode="bar")

# %%
# Similarly, the aggregate information can be visualized using a heatmap or a clustermap.
cr.pl.cluster_fates(adata, mode="heatmap")
cr.pl.cluster_fates(adata, mode="clustermap")

# %%
# Violin plots are helpful to visualize the distribution of fate probabilities per cluster.
# It is also possible to restrict this plot only to a subset of clusters using the ``clusters`` parameter.
cr.pl.cluster_fates(adata, mode="violin", cluster_key="clusters")

# %%
# We can also plot the PAGA graph, using an UMAP embedding while visualizing the aggregate absorption probabilities
# as pie charts.
#
# However, using :func:`scvelo.tl.paga` and :mod:`cellrank`'s initial and terminal state probabilities, we can also
# visualize a directed version of PAGA, see :ref:`sphx_glr_auto_examples_plotting_plot_directed_paga.py`.
cr.pl.cluster_fates(adata, mode="paga_pie", basis="umap", cluster_key="clusters")

# %%
# Lastly, we can visualize the absorption probabilities in PAGA graph by coloring the nodes.
cr.pl.cluster_fates(adata, mode="paga", legend_loc="on data", basis="umap")



from typing import Any, Dict, List, Tuple, Union, Callable, Iterable, Optional

import wrapt
from types import MappingProxyType
from functools import wraps, update_wrapper
from contextlib import contextmanager
from multiprocessing import cpu_count

from anndata import AnnData
from cellrank import logging as logg
from anndata.utils import make_index_unique
from cellrank.ul._docs import d

import numpy as np
from scipy.sparse import issparse, spmatrix


def _check_collection(
    adata: AnnData,
    needles: Iterable[str],
    attr_name: str,
    key_name: str = "Gene",
    use_raw: bool = False,
    raise_exc: bool = True,
) -> List[str]:
    """
    Check if given collection contains all the keys.

    Parameters
    ----------
    adata: :class:`anndata.AnnData`
        Annotated data object.
    needles
        Keys to check.
    attr_name
        Attribute of ``adata`` where the needles are stored.
    key_name
        Pretty name of the key which will be displayed when error is found.
    use_raw
        Whether to access ``adata.raw`` or just ``adata``.

    Returns
    -------
    Nothing, but raises and :class:`KeyError` if one of the needles is not found.
    """
    adata_name = "adata"

    if use_raw and adata.raw is None:
        logg.warning(
            "Argument `use_raw` was set to `True`, but no `raw` attribute is found. Ignoring"
        )
        use_raw = False
    if use_raw:
        adata_name = "adata.raw"
        adata = adata.raw

    haystack, res = getattr(adata, attr_name), []
    for needle in needles:
        if needle not in haystack:
            if raise_exc:
                raise KeyError(
                    f"{key_name} `{needle}` not found in `{adata_name}.{attr_name}`."
                )
        else:
            res.append(needle)

    return res


def _get_n_cores(n_cores: Optional[int], n_jobs: Optional[int]) -> int:
    """
    Make number of cores a positive integer.

    Parameters
    ----------
    n_cores
        Number of cores to use.
    n_jobs.
        Number of jobs. This is just used to determine if the collection is a singleton.
        If `1`, always returns `1`.

    Returns
    -------
    Positive integer corresponding to how many cores to use.
    """
    if n_cores == 0:
        raise ValueError("Number of cores cannot be `0`.")
    if n_jobs == 1 or n_cores is None:
        return 1
    if n_cores < 0:
        return cpu_count() + 1 + n_cores

    return n_cores


def _minmax(
    data: np.ndarray, perc: Optional[Tuple[float, float]] = None
) -> Tuple[float, float]:
    """
    Return minimum and maximum value of the data.

    Parameters
    ----------
    data
        Values for which to return the minimum and maximum.
    perc
        If not `None`, clip the values by the percentiles before getting the result.

    Returns
    -------
    Minimum and maximum values, respectively.
    """
    if perc is not None:
        data = np.clip(data, *np.percentile(data, sorted(perc)))

    return float(np.nanmin(data)), float(np.nanmax(data))


def _modify_neigh_key(key: Optional[str]) -> str:
    if key in (None, "connectivities", "distances"):
        return "neighbors"

    if key.endswith("_connectivities"):
        key = key[:-15]
    elif key.endswith("_distances"):
        key = key[:-10]

    return key


def _get_neighs(
    adata: AnnData, mode: str = "distances", key: Optional[str] = None
) -> Union[np.ndarray, spmatrix]:
    if key is None:
        res = _read_graph_data(adata, mode)  # legacy behavior
    else:
        try:
            res = _read_graph_data(adata, key)
            assert isinstance(res, (np.ndarray, spmatrix))
        except (KeyError, AssertionError):
            res = _read_graph_data(adata, f"{_modify_neigh_key(key)}_{mode}")

    if not isinstance(res, (np.ndarray, spmatrix)):
        raise TypeError(
            f"Expected to find `numpy.ndarray` or `scipy.sparse.spmatrix`, found `{type(res)}`."
        )

    return res


def _has_neighs(adata: AnnData, key: Optional[str] = None) -> bool:
    return _modify_neigh_key(key) in adata.uns.keys()


def _get_neighs_params(adata: AnnData, key: str = "neighbors") -> Dict[str, Any]:
    return adata.uns.get(key, {}).get("params", {})


def _read_graph_data(adata: AnnData, key: str) -> Union[np.ndarray, spmatrix]:
    """
    Read graph data from :mod:`anndata`.

    Parameters
    ----------
    adata
        Annotated data object.
    key
        Key in :attr:`anndata.AnnData.obsp`.

    Returns
    -------
    The graph data.
    """
    if key in adata.obsp:
        return adata.obsp[key]

    raise KeyError(f"Unable to find data in `adata.obsp[{key!r}]`.")


def valuedispatch(func):
    """Dispatch a function based on the first value."""
    registry = {}

    def dispatch(value):
        return registry.get(value, func)

    def register(value, func=None):
        if func is None:
            return lambda f: register(value, f)
        registry[value] = func
        return func

    @wraps(func)
    def wrapper(*args, **kwargs):
        return dispatch(args[0])(*args[1:], **kwargs)

    wrapper.register = register
    wrapper.dispatch = dispatch
    wrapper.registry = MappingProxyType(registry)

    update_wrapper(wrapper, func)

    return wrapper


def _densify_squeeze(x: Union[spmatrix, np.ndarray], dtype=np.float32) -> np.ndarray:
    if issparse(x):
        x = x.toarray()
    # use np.array instead of asarray to create a copy
    x = np.array(x, dtype=dtype)
    if x.ndim == 2 and x.shape[1] == 1:
        x = np.squeeze(x, axis=1)

    return x


@contextmanager
@d.dedent
def _gene_symbols_ctx(
    adata: AnnData,
    *,
    key: Optional[str] = None,
    use_raw: bool = False,
    make_unique: bool = False,
) -> AnnData:
    """
    Set gene names from a column in :attr:`anndata.AnnData.var`.

    Parameters
    ----------
    %(adata)s
    key
        Key in :attr:`anndata.AnnData.var` where the gene symbols are stored. If `None`, this operation is a no-op.
    use_raw
        Whether to change the gene names in :attr:`anndata.AnnData.raw`.
    make_unique
        Whether to make the newly assigned gene names unique.
    Yields
    ------
    The same ``adata`` with modified :attr:`anndata.AnnData.var_names`, depending on ``use_raw``.
    """

    def key_present() -> bool:
        if use_raw:
            if adata.raw is None:
                raise AttributeError(
                    "No `.raw` attribute found. Try specifying `use_raw=False`."
                )
            return key in adata.raw.var
        return key in adata.var

    if key is None:
        yield adata
    elif not key_present():
        raise KeyError(
            f"Unable to find gene symbols in `adata.{'raw.' if use_raw else ''}var[{key!r}]`."
        )
    else:
        adata_orig = adata
        if use_raw:
            adata = adata.raw

        var_names = adata.var_names.copy()
        try:
            # TODO(michalk8): doesn't update varm (niche)
            adata.var.index = (
                make_index_unique(adata.var[key]) if make_unique else adata.var[key]
            )
            yield adata_orig
        finally:
            # in principle we assume the callee doesn't change the index
            # otherwise, would need to check whether it has been changed and add an option to determine what to do
            adata.var.index = var_names


@wrapt.decorator
def _genesymbols(
    wrapped: Callable[..., Any], instance: Any, args: Any, kwargs: Any
) -> Any:
    with _gene_symbols_ctx(
        args[0],
        key=kwargs.pop("gene_symbols", None),
        make_unique=True,
        use_raw=kwargs.get("use_raw", False),
    ):
        return wrapped(*args, **kwargs)



from typing import Tuple, Optional

import os
import warnings
from sys import version_info
from pathlib import Path
from filelock import FileLock

from cellrank.ul.models import GAM, GAMR, SKLearnModel

os.environ["NUMBA_NUM_THREADS"] = "2"

import pytest
from _helpers import create_model

import scanpy as sc
import cellrank as cr
from anndata import AnnData
from cellrank.tl.kernels import VelocityKernel, ConnectivityKernel
from cellrank.tl.estimators import GPCCA, CFLARE

import numpy as np
from numba.core.errors import NumbaPerformanceWarning

import matplotlib

_adata_small = sc.read("tests/_ground_truth_adatas/adata_50.h5ad")
_adata_medium = sc.read("tests/_ground_truth_adatas/adata_100.h5ad")
_adata_large = sc.read("tests/_ground_truth_adatas/adata_200.h5ad")


def pytest_sessionstart(session) -> None:
    matplotlib.use("Agg")
    matplotlib.rcParams["figure.max_open_warning"] = 0
    np.random.seed(42)

    # https://github.com/theislab/cellrank/issues/683
    warnings.simplefilter("ignore", NumbaPerformanceWarning)


# removes overly verbose and useless logging errors for rpy2
# see: https://github.com/pytest-dev/pytest/issues/5502#issuecomment-647157873
def pytest_sessionfinish(session, exitstatus) -> None:
    import logging

    loggers = [logging.getLogger()] + list(logging.Logger.manager.loggerDict.values())
    for logger in loggers:
        handlers = getattr(logger, "handlers", [])
        for handler in handlers:
            logger.removeHandler(handler)


def _create_cflare(*, backward: bool = False) -> Tuple[AnnData, CFLARE]:
    adata = _adata_medium.copy()

    sc.tl.paga(adata, groups="clusters")

    vk = VelocityKernel(adata, backward=backward).compute_transition_matrix(
        softmax_scale=4
    )
    ck = ConnectivityKernel(adata, backward=backward).compute_transition_matrix()
    final_kernel = 0.8 * vk + 0.2 * ck

    mc = CFLARE(final_kernel)
    final_kernel.write_to_adata()

    mc.compute_eigendecomposition()
    mc.compute_terminal_states(use=2, method="kmeans")
    mc.compute_absorption_probabilities(use_petsc=False)
    mc.compute_lineage_drivers(cluster_key="clusters", use_raw=False)

    assert adata is mc.adata
    np.testing.assert_allclose(mc.absorption_probabilities.X.sum(1), 1.0, rtol=1e-6)

    return adata, mc


def _create_gpcca(*, backward: bool = False) -> Tuple[AnnData, GPCCA]:
    adata = _adata_medium.copy()

    sc.tl.paga(adata, groups="clusters")

    vk = VelocityKernel(adata, backward=backward).compute_transition_matrix(
        softmax_scale=4
    )
    ck = ConnectivityKernel(adata, backward=backward).compute_transition_matrix()
    final_kernel = 0.8 * vk + 0.2 * ck

    mc = GPCCA(final_kernel)
    final_kernel.write_to_adata()

    mc.compute_eigendecomposition()
    mc.compute_schur(method="krylov")
    mc.compute_macrostates(n_states=2)
    mc.set_terminal_states_from_macrostates()
    mc.compute_absorption_probabilities(use_petsc=False)
    mc.compute_lineage_drivers(cluster_key="clusters", use_raw=False)

    assert adata is mc.adata
    np.testing.assert_allclose(mc.absorption_probabilities.X.sum(1), 1.0, rtol=1e-6)

    return adata, mc


def _create_gamr_model(_adata: AnnData) -> Optional[GAMR]:
    try:
        m = GAMR(_adata)
        m.prepare(_adata.var_names[0], "0").fit()
        m.predict(level=0.95)
        return m
    except Exception:
        return None


@pytest.fixture
def adata() -> AnnData:
    return _adata_small.copy()


@pytest.fixture
def adata_large() -> AnnData:
    return _adata_large.copy()


@pytest.fixture
def adata_cflare_fwd(
    adata_cflare=_create_cflare(backward=False),
) -> Tuple[AnnData, CFLARE]:
    adata, cflare = adata_cflare
    return adata.copy(), cflare


@pytest.fixture
def adata_gpcca_fwd(adata_gpcca=_create_gpcca(backward=False)) -> Tuple[AnnData, GPCCA]:
    adata, gpcca = adata_gpcca
    return adata.copy(), gpcca


@pytest.fixture
def adata_gpcca_bwd(adata_gpcca=_create_gpcca(backward=True)) -> Tuple[AnnData, GPCCA]:
    adata, gpcca = adata_gpcca
    return adata.copy(), gpcca


@pytest.fixture
def adata_cflare(adata_cflare=_create_cflare(backward=False)) -> AnnData:
    return adata_cflare[0].copy()


@pytest.fixture(scope="session")
def adata_gamr(adata_cflare=_create_cflare(backward=False)) -> AnnData:
    return adata_cflare[0].copy()


@pytest.fixture(scope="session")
def gamr_model(
    adata_gamr: AnnData, tmp_path_factory: Path, worker_id: str
) -> Optional[GAMR]:
    model = None

    if version_info[:2] <= (3, 6):
        pytest.skip("Pickling of Enums doesn't work in Python3.6.")
    elif worker_id == "master":
        model = _create_gamr_model(adata_gamr)
    else:
        root_tmp_dir = tmp_path_factory.getbasetemp().parent
        fn = root_tmp_dir / "model.pickle"

        with FileLock(f"{fn}.lock"):
            if fn.is_file():
                model = GAMR.read(fn)
            else:
                model = _create_gamr_model(adata_gamr)
                if model is not None:
                    model.write(fn)

    if model is None:
        pytest.skip("Unable to create `cellrank.ul.models.GAMR`.")

    return model


@pytest.fixture
def pygam_model(adata_cflare: AnnData) -> GAM:
    m = GAM(adata_cflare)
    m.prepare(adata_cflare.var_names[0], "0").fit()
    m.predict()
    m.confidence_interval()

    return m


@pytest.fixture
def sklearn_model(adata_cflare: AnnData) -> SKLearnModel:
    m = create_model(adata_cflare)
    assert isinstance(m, SKLearnModel), m

    m.prepare(adata_cflare.var_names[0], "0").fit()
    m.predict()
    m.confidence_interval()

    return m


@pytest.fixture
def lineage():
    x = cr.tl.Lineage(
        np.array(
            [
                [1.23459664e-01, 1.29965675e-01, 1.92828002e-01, 9.39402664e-01],
                [1.05635239e00, 4.45833459e-01, 2.29080759e00, 1.90132652e00],
                [6.77880737e-02, 4.97556864e-02, 1.18428661e00, 2.02318999e-01],
                [4.87500398e-01, 1.00657498e00, 2.20834882e-02, 5.03008905e-01],
                [6.27190917e00, 7.27864781e00, 1.03978903e00, 1.55903460e01],
                [3.85149269e-01, 3.54765380e-01, 1.77871487e-01, 8.22138648e-02],
                [7.06618729e00, 1.33133671e01, 1.44904591e00, 5.79813391e00],
                [8.18005744e-02, 5.36844933e-01, 1.86646162e00, 2.41141727e00],
                [1.44892035e-01, 2.34036215e-01, 6.32392890e-01, 1.13211403e-02],
                [2.44926466e-01, 2.50293183e-01, 1.77540208e-01, 3.27240144e-01],
            ]
        ),
        names=["foo", "bar", "baz", "quux"],
    )
    return x / x.sum(1)


@pytest.fixture
def kernel(adata_large: AnnData):
    vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
    ck = ConnectivityKernel(adata_large).compute_transition_matrix()
    return (0.8 * vk + 0.2 * ck).compute_transition_matrix()


@pytest.fixture(scope="session")
def test_matrix_1() -> np.ndarray:
    """
    Parameters
    ----------

    Returns
    -------
    :class:`numpy.ndarray`
        Row-normalized transition matrix. This matrix is
        - connected
        - irreducible
        - not reversible
    """

    # fmt: off
    p = np.array([
        # 0.   1.   2.   3.   4.   5.   6.   7.   8.   9.   10.  11.
        [0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 0
        [0.2, 0.0, 0.6, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 1
        [0.6, 0.2, 0.0, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 2

        [0.0, 0.05, 0.05, 0.0, 0.6, 0.3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 3
        [0.0, 0.0, 0.0, 0.25, 0.0, 0.25, 0.4, 0.0, 0.0, 0.1, 0.0, 0.0],  # 4
        [0.0, 0.0, 0.0, 0.25, 0.25, 0.0, 0.1, 0.0, 0.0, 0.4, 0.0, 0.0],  # 5

        [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.7, 0.2, 0.0, 0.0, 0.0],  # 6
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0, 0.8, 0.0, 0.0, 0.0],  # 7
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0],  # 8

        [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.0, 0.0, 0.0, 0.7, 0.2],  # 9
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0, 0.8],  # 10
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0],  # 11
    ])
    # fmt: on

    return p


@pytest.fixture(scope="session")
def test_matrix_2() -> np.ndarray:
    """
    Parameters
    ----------

    Returns
    -------
    :class:`numpy.ndarray`
        Row-normalized transition matrix. This matrix is
        - connected
        - not irreducible (1 recurrent, 1 transient class)
        - not reversible
    """
    # fmt: off
    p = np.array([
        # 0.   1.   2.   3.   4.   5.   6.   7.   8.   9.   10.  11.  12.  13.
        [0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 0
        [0.2, 0.0, 0.6, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 1
        [0.6, 0.2, 0.0, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 2

        [0.0, 0.05, 0.05, 0.0, 0.6, 0.3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 3
        [0.0, 0.0, 0.0, 0.25, 0.0, 0.25, 0.4, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0],  # 4
        [0.0, 0.0, 0.0, 0.25, 0.25, 0.0, 0.1, 0.0, 0.0, 0.3, 0.0, 0.0, 0.1, 0.0],  # 5

        [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.7, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0],  # 6
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0, 0.8, 0.0, 0.0, 0.0, 0.0, 0.0],  # 7
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 8

        [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.0, 0.0, 0.0, 0.7, 0.2, 0.0, 0.0],  # 9
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0, 0.8, 0.0, 0.0],  # 10
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0, 0.0, 0.0],  # 11

        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, 0.4],  # 12
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.3, 0.7],  # 13
    ])
    # fmt: on

    return p


@pytest.fixture(scope="session")
def test_matrix_3() -> np.ndarray:
    """
    Parameters
    ----------

    Returns
    -------
    :class:`numpy.ndarray`
        Row-normalized transition matrix. This matrix is
        - not connected
        - not irreducible (2 recurrent classes)
        - not reversible
    """

    # fmt: off
    p = np.array([
        # 0.   1.   2.   3.   4.   5.   6.   7.   8.   9.   10.  11.  12.  13.
        [0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 0
        [0.2, 0.0, 0.6, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 1
        [0.6, 0.2, 0.0, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 2

        [0.0, 0.05, 0.05, 0.0, 0.6, 0.3, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 3
        [0.0, 0.0, 0.0, 0.25, 0.0, 0.25, 0.4, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0],  # 4
        [0.0, 0.0, 0.0, 0.25, 0.25, 0.0, 0.1, 0.0, 0.0, 0.4, 0.0, 0.0, 0.0, 0.0],  # 5

        [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.7, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0],  # 6
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0, 0.8, 0.0, 0.0, 0.0, 0.0, 0.0],  # 7
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],  # 8

        [0.0, 0.0, 0.0, 0.0, 0.05, 0.05, 0.0, 0.0, 0.0, 0.0, 0.7, 0.2, 0.0, 0.0],  # 9
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0, 0.8, 0.0, 0.0],  # 10
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8, 0.2, 0.0, 0.0, 0.0],  # 11

        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, 0.4],  # 12
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.3, 0.7],  # 13
    ])
    # fmt: on

    return p


@pytest.fixture(scope="session")
def test_matrix_4() -> np.ndarray:
    """
    Parameters
    ----------

    Returns
    -------
    :class:`numpy.ndarray`
        Symmetric matrix. Not a transition matrix.
    """

    # fmt: off
    p = np.array(
        [
            # 0.   1.   2.   3.
            [0.0, 0.8, 0.2, 0.0],  # 0
            [0.8, 0.0, 0.6, 0.2],  # 1
            [0.2, 0.6, 0.0, 0.2],  # 2
            [0.0, 0.2, 0.2, 0.0],  # 3
        ]
    )
    # fmt: on

    return p



from typing import Any, Optional


class ImportErrorMixin:
    """
    Mixin class which always raises :class:`ImportError`.

    Subclasses can modify the message by overriding `__import_error_message__`.

    Parameters
    ----------
    args
        Ignored.
    kwargs
        Ignored.

    Raises
    ------
    ImportError
        Always.
    """

    __import_error_message__ = "Unable to import the class."

    def __init__(self, *args: Any, **kwargs: Any):
        raise ImportError(self.__import_error_message__) from self.__error__

    def __init_subclass__(cls, error: Optional[Exception] = None, **kwargs: Any):
        super().__init_subclass__(**kwargs)
        cls.__error__ = error



from typing import Any, Dict, Tuple, Union, Callable, Optional, Sequence
from typing_extensions import Literal

from copy import deepcopy
from types import MappingProxyType
from pathlib import Path

from anndata import AnnData
from cellrank import logging as logg
from cellrank._key import Key
from cellrank.ul._docs import d
from cellrank.tl._utils import save_fig, _unique_order_preserving
from cellrank.ul._utils import _read_graph_data

import numpy as np
import pandas as pd
from scipy.sparse import issparse, spmatrix
from pandas.api.types import is_categorical_dtype

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.colors import ListedColormap
from matplotlib.patches import ArrowStyle, FancyArrowPatch
from matplotlib.collections import LineCollection
from mpl_toolkits.axes_grid1 import make_axes_locatable


@d.dedent
def graph(
    data: Union[AnnData, np.ndarray, spmatrix],
    graph_key: Optional[str] = None,
    ixs: Optional[Union[range, np.array]] = None,
    layout: Union[str, Dict[str, Any], Callable[..., np.ndarray]] = "umap",
    keys: Sequence[
        Union[str, Literal["incoming", "outgoing", "self_loops"]]
    ] = "incoming",
    keylocs: Union[str, Sequence[str]] = "uns",
    node_size: float = 400,
    labels: Optional[Union[Sequence[str], Sequence[Sequence[str]]]] = None,
    top_n_edges: Optional[Union[int, Tuple[int, bool, str]]] = None,
    self_loops: bool = True,
    self_loop_radius_frac: Optional[float] = None,
    filter_edges: Optional[Tuple[float, float]] = None,
    edge_reductions: Union[Callable, Sequence[Callable]] = np.sum,
    edge_reductions_restrict_to_ixs: Optional[Union[range, np.ndarray]] = None,
    edge_weight_scale: float = 10,
    edge_width_limit: Optional[float] = None,
    edge_alpha: float = 1.0,
    edge_normalize: bool = False,
    edge_use_curved: bool = True,
    arrows: bool = True,
    font_size: int = 12,
    font_color: str = "black",
    color_nodes: bool = True,
    cat_cmap: ListedColormap = cm.Set3,
    cont_cmap: ListedColormap = cm.viridis,
    legend_loc: Optional[str] = "best",
    title: Optional[Union[str, Sequence[Optional[str]]]] = None,
    figsize: Optional[Tuple[float, float]] = None,
    dpi: Optional[int] = None,
    save: Optional[Union[str, Path]] = None,
    layout_kwargs: Dict[str, Any] = MappingProxyType({}),
) -> None:
    """
    Plot a graph, visualizing incoming and outgoing edges or self-transitions.

    This is a utility function to look in more detail at the transition matrix in areas of interest, e.g. around an
    endpoint of development. This function is meant to visualise a small subset of nodes (~100-500) and the most likely
    transitions between them. Note that limiting edges visualized using ``top_n_edges`` will speed things up,
    as well as reduce the visual clutter.

    Parameters
    ----------
    data
        The graph data to be plotted.
    graph_key
        Key in :attr:`anndata.AnnData.obsp` where the graph is stored.
        Only used when ``data`` is :class:`anndata.Anndata` object.
    ixs
        Subset of indices of the graph to visualize.
    layout
        Layout to use for graph drawing.

        - If :class:`str`, search for embedding in :attr:`anndata.AnnData.obsm` ``['X_{layout}']``.
          Use ``layout_kwargs={'components': [0, 1]}`` to select components.
        - If :class:`dict`, keys should be values in interval `[0, len(ixs))`
          and values `(x, y)` pairs corresponding to node positions.
    keys
        Keys in :attr:`anndata.AnnData.obs`, :attr:`anndata.AnnData.obsm` or :attr:`anndata.AnnData.obsm`
        used to color the nodes.

        If `'incoming'`, `'outgoing'` or `'self_loops'`, visualize reduction (see ``edge_reductions``)
        for each node based on incoming or outgoing edges, respectively.
    keylocs
        Locations of ``keys``. Can be any attribute of ``data`` if it's :class:`anndata.AnnData` object.
    node_size
        Size of the nodes.
    labels
        Labels of the nodes.
    top_n_edges
        Either top N outgoing edges in descending order or a tuple
        ``(top_n_edges, in_ascending_order, {'incoming', 'outgoing'})``. If `None`, show all edges.
    self_loops
        Whether visualize self transitions and also to consider them in ``top_n_edges``.
    self_loop_radius_frac
        Fraction of a unit circle to visualize self transitions. If `None`, use ``node_size / 1000``.
    filter_edges
        Whether to remove all edges not in `[min, max]` interval.
    edge_reductions
        Aggregation function to use when coloring nodes by edge weights.
    edge_reductions_restrict_to_ixs
        Whether to use the full graph when calculating the ``edge_reductions`` or just use the nodes
        marked by the ``ixs`` and this parameter. If `None`, it's the same as ``ixs``.
    edge_weight_scale
        Number by which to scale the width of the edges. Useful when the weights are small.
    edge_width_limit
        Upper bound for the width of the edges. Useful when weights are unevenly distributed.
    edge_alpha
        Alpha channel value for edges and arrows.
    edge_normalize
        If `True`, normalize edges to `[0, 1]` interval prior to applying any scaling or truncation.
    edge_use_curved
        If `True`, use curved edges. This can improve visualization at a small performance cost.
    arrows
        Whether to show the arrows. Setting this to `False` may dramatically speed things up.
    font_size
        Font size for node labels.
    font_color
        Label color of the nodes.
    color_nodes
        Whether to color the nodes
    cat_cmap
        Categorical colormap used when ``keys`` contain categorical variables.
    cont_cmap
        Continuous colormap used when ``keys`` contain continuous variables.
    legend_loc
        Location of the legend.
    title
        Title of the figure(s), one for each ``key``.
    %(plotting)s
    layout_kwargs
        Keyword arguments for ``layout``.

    Returns
    -------
    %(just_plots)s
    """
    import networkx as nx

    def plot_arrows(curves, G, pos, ax, edge_weight_scale):
        for line, (edge, val) in zip(curves, G.edges.items()):
            if edge[0] == edge[1]:
                continue

            mask = (~np.isnan(line)).all(axis=1)
            line = line[mask, :]
            if not len(line):  # can be all NaNs
                continue

            line = line.reshape((-1, 2))
            X, Y = line[:, 0], line[:, 1]

            node_start = pos[edge[0]]
            # reverse
            if np.where(np.isclose(node_start - line, [0, 0]).all(axis=1))[0][0]:
                X, Y = X[::-1], Y[::-1]

            mid = len(X) // 2
            posA, posB = zip(X[mid : mid + 2], Y[mid : mid + 2])

            arrow = FancyArrowPatch(
                posA=posA,
                posB=posB,
                # we clip because too small values
                # cause it to crash
                arrowstyle=ArrowStyle.CurveFilledB(
                    head_length=np.clip(
                        val["weight"] * edge_weight_scale * 4,
                        _min_edge_weight,
                        edge_width_limit,
                    ),
                    head_width=np.clip(
                        val["weight"] * edge_weight_scale * 2,
                        _min_edge_weight,
                        edge_width_limit,
                    ),
                ),
                color="k",
                zorder=float("inf"),
                alpha=edge_alpha,
                linewidth=0,
            )
            ax.add_artist(arrow)

    def normalize_weights():
        weights = np.array([v["weight"] for v in G.edges.values()])
        minn = np.min(weights)
        weights = (weights - minn) / (np.max(weights) - minn)
        for v, w in zip(G.edges.values(), weights):
            v["weight"] = w

    def remove_top_n_edges():
        if top_n_edges is None:
            return

        if isinstance(top_n_edges, (tuple, list)):
            to_keep, ascending, group_by = top_n_edges
        else:
            to_keep, ascending, group_by = top_n_edges, False, "out"

        if group_by not in ("incoming", "outgoing"):
            raise ValueError(
                f"Argument `groupby` in `top_n_edges` must be either `'incoming`' or `'outgoing'`, found `{group_by}`."
            )

        source, target = zip(*G.edges)
        weights = [v["weight"] for v in G.edges.values()]
        tmp = pd.DataFrame({"outgoing": source, "incoming": target, "w": weights})

        if not self_loops:
            # remove self loops
            tmp = tmp[tmp["incoming"] != tmp["outgoing"]]

        to_keep = set(
            map(
                tuple,
                tmp.groupby(group_by)
                .apply(
                    lambda g: g.sort_values("w", ascending=ascending).take(
                        range(min(to_keep, len(g)))
                    )
                )[["outgoing", "incoming"]]
                .values,
            )
        )

        for e in list(G.edges):
            if e not in to_keep:
                G.remove_edge(*e)

    def remove_low_weight_edges():
        if filter_edges is None or filter_edges == (None, None):
            return

        minn, maxx = filter_edges
        minn = minn if minn is not None else -np.inf
        maxx = maxx if maxx is not None else np.inf

        for e, attr in list(G.edges.items()):
            if attr["weight"] < minn or attr["weight"] > maxx:
                G.remove_edge(*e)

    _min_edge_weight = 0.00001

    if edge_width_limit is None:
        logg.debug("Not limiting width of edges")
        edge_width_limit = float("inf")

    if self_loop_radius_frac is None:
        self_loop_radius_frac = (
            node_size / 2000 if node_size >= 200 else node_size / 1000
        )
        logg.debug(f"Setting self loop radius fraction to `{self_loop_radius_frac}`")

    if isinstance(keys, str):
        keys = [keys]
    keys = _unique_order_preserving(keys)

    if not isinstance(keylocs, (tuple, list)):
        keylocs = [keylocs] * len(keys)
    elif len(keylocs) == 1:
        keylocs = keylocs * 3
    elif all(map(lambda k: k in ("incoming", "outgoing", "self_loops"), keys)):
        # don't care about keylocs since they are irrelevant
        logg.debug("Ignoring key locations")
        keylocs = [None] * len(keys)

    if not isinstance(edge_reductions, (tuple, list)):
        edge_reductions = [edge_reductions] * len(keys)
    if not all(map(callable, edge_reductions)):
        raise ValueError("Not all `edge_reductions` functions are callable.")

    if not isinstance(labels, (tuple, list)):
        labels = [labels] * len(keys)
    elif not len(labels):
        labels = [None] * len(keys)
    elif not isinstance(labels[0], (tuple, list)):
        labels = [labels] * len(keys)

    if len(keys) != len(labels):
        raise ValueError(
            f"`Keys` and `labels` must be of the same shape, found `{len(keys)}` and `{len(labels)}`."
        )
    if title is None or isinstance(title, str):
        title = [title] * len(keys)
    if len(title) != len(keys):
        raise ValueError(
            f"`Titles` and `keys` must be of the same shape, found `{len(title)}` and `{len(keys)}`."
        )

    if isinstance(data, AnnData):
        if graph_key is None:
            raise ValueError(
                "Argument `graph_key` cannot be `None` when `data` is `anndata.Anndata` object."
            )
        gdata = _read_graph_data(data, graph_key)
    elif isinstance(data, (np.ndarray, spmatrix)):
        gdata = data
    else:
        raise TypeError(
            f"Expected argument `data` to be one of `anndata.AnnData`, `numpy.ndarray`, `scipy.sparse.spmatrix`, "
            f"found `{type(data).__name__!r}`."
        )
    is_sparse = issparse(gdata)

    gdata_full = gdata
    if ixs is not None:
        gdata = gdata[ixs, :][:, ixs]
    else:
        ixs = list(range(gdata.shape[0]))

    if edge_reductions_restrict_to_ixs is None:
        edge_reductions_restrict_to_ixs = ixs

    start = logg.info("Creating graph")
    G = (
        nx.from_scipy_sparse_matrix(gdata, create_using=nx.DiGraph)
        if is_sparse
        else nx.from_numpy_array(gdata, create_using=nx.DiGraph)
    )

    remove_low_weight_edges()
    remove_top_n_edges()
    if edge_normalize:
        normalize_weights()
    logg.info("    Finish", time=start)

    # do NOT recreate the graph, for the edge reductions
    # gdata = nx.to_numpy_array(G)

    if figsize is None:
        figsize = (12, 8 * len(keys))

    fig, axes = plt.subplots(nrows=len(keys), ncols=1, figsize=figsize, dpi=dpi)
    if not isinstance(axes, np.ndarray):
        axes = np.array([axes])
    axes = np.ravel(axes)

    if isinstance(layout, str):
        if f"X_{layout}" not in data.obsm:
            raise KeyError(f"Unable to find embedding `'X_{layout}'` in `adata.obsm`.")
        components = layout_kwargs.get("components", [0, 1])
        if len(components) != 2:
            raise ValueError(
                f"Components in `layout_kwargs` must be of length `2`, found `{len(components)}`."
            )
        emb = data.obsm[f"X_{layout}"][:, components]
        pos = {i: emb[ix, :] for i, ix in enumerate(ixs)}
        logg.info(f"Embedding graph using `{layout!r}` layout")
    elif isinstance(layout, dict):
        rng = range(len(ixs))
        for k, v in layout.items():
            if k not in rng:
                raise ValueError(
                    f"Key in `layout` must be in `range(len(ixs))`, found `{k}`."
                )
            if len(v) != 2:
                raise ValueError(
                    f"Value in `layout` must be a `tuple` or a `list` of length 2, found `{len(v)}`."
                )
        pos = layout
        logg.debug("Using precomputed layout")
    elif callable(layout):
        start = logg.info(f"Embedding graph using `{layout.__name__!r}` layout")
        pos = layout(G, **layout_kwargs)
        logg.info("    Finish", time=start)
    else:
        raise TypeError(
            f"Argument `layout` must be either a `string`, "
            f"a `dict` or a `callable`, found `{type(layout)}`."
        )

    curves, lc = None, None
    if edge_use_curved:
        try:
            from cellrank.pl._utils import _curved_edges

            logg.debug("Creating curved edges")
            curves = _curved_edges(G, pos, self_loop_radius_frac, polarity="directed")
            lc = LineCollection(
                curves,
                colors="black",
                linewidths=np.clip(
                    np.ravel([v["weight"] for v in G.edges.values()])
                    * edge_weight_scale,
                    0,
                    edge_width_limit,
                ),
                alpha=edge_alpha,
            )
        except ImportError:
            logg.error(
                "Unable to show curved edges. Please install `bezier` as `pip install bezier` or "
                "use `edge_use_curved=False`"
            )

    for ax, keyloc, title, key, labs, er in zip(
        axes, keylocs, title, keys, labels, edge_reductions
    ):
        label_col = {}  # dummy value

        if key in ("incoming", "outgoing", "self_loops"):
            if key in ("incoming", "outgoing"):
                axis = int(key == "outgoing")
                tmp_data = (
                    gdata_full[ixs, :][:, edge_reductions_restrict_to_ixs]
                    if axis
                    else gdata_full[edge_reductions_restrict_to_ixs, :][:, ixs]
                )
                vals = er(tmp_data, axis=axis)
            else:
                vals = gdata.diagonal() if is_sparse else np.diag(gdata)

            if issparse(vals):
                vals = vals.toarray()
            vals = np.squeeze(np.array(vals))

            assert len(pos) == len(vals), (
                f"Expected reduction values to be of length `{len(pos)}`, "
                f"found `{len(vals)}`."
            )

            node_v = dict(zip(pos.keys(), vals))
        else:
            label_col = getattr(data, keyloc)
            if key in label_col:
                node_v = dict(zip(pos.keys(), label_col[key]))
            else:
                raise RuntimeError(f"Key `{key!r}` not found in `adata.{keyloc}`.")

        if labs is not None:
            if len(labs) != len(pos):
                raise RuntimeError(
                    f"Number of labels ({len(labels)}) and nodes ({len(pos)}) mismatch."
                )
            nx.draw_networkx_labels(
                G,
                pos,
                labels=labs if isinstance(labs, dict) else dict(zip(pos.keys(), labs)),
                ax=ax,
                font_color=font_color,
                font_size=font_size,
            )

        if lc is not None and curves is not None:
            ax.add_collection(deepcopy(lc))  # copying necessary
            if arrows:
                plot_arrows(curves, G, pos, ax, edge_weight_scale)
        else:
            nx.draw_networkx_edges(
                G,
                pos,
                width=[
                    np.clip(
                        v["weight"] * edge_weight_scale,
                        _min_edge_weight,
                        edge_width_limit,
                    )
                    for _, v in G.edges.items()
                ],
                alpha=edge_alpha,
                edge_color="black",
                arrows=True,
                arrowstyle="-|>",
            )

        if key in label_col and is_categorical_dtype(label_col[key]):
            values = label_col[key]
            if keyloc in ("obs", "obsm"):
                values = values[ixs]
            categories = values.cat.categories
            color_key = Key.uns.colors(key)
            if color_key in data.uns:
                mapper = dict(zip(categories, data.uns[color_key]))
            else:
                mapper = dict(
                    zip(categories, map(cat_cmap.get, range(len(categories))))
                )

            colors = []
            seen = set()

            for v in values:
                colors.append(mapper[v])
                seen.add(v)

            nodes_kwargs = dict(cmap=cat_cmap, node_color=colors)  # noqa
            if legend_loc not in ("none", None):
                x, y = pos[0]
                for label in sorted(seen):
                    ax.plot([x], [y], label=label, color=mapper[label])
                ax.legend(loc=legend_loc)
        else:
            values = list(node_v.values())
            vmin, vmax = np.min(values), np.max(values)
            nodes_kwargs = dict(  # noqa
                cmap=cont_cmap, node_color=values, vmin=vmin, vmax=vmax
            )

            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="1.5%", pad=0.05)
            _ = mpl.colorbar.ColorbarBase(
                cax,
                cmap=cont_cmap,
                norm=mpl.colors.Normalize(vmin=vmin, vmax=vmax),
                ticks=np.linspace(vmin, vmax, 10),
            )

        if color_nodes is False:
            nodes_kwargs = {}

        nx.draw_networkx_nodes(G, pos, node_size=node_size, ax=ax, **nodes_kwargs)

        ax.set_title(key if title is None else title)
        ax.axis("off")

    if save is not None:
        save_fig(fig, save)



from typing import Any, Dict, Tuple, Union, Mapping, Optional
from typing_extensions import Literal

from types import MappingProxyType
from tqdm.auto import tqdm

import scanpy as sc
from anndata import AnnData
from cellrank import logging as logg
from cellrank.ul._docs import d, inject_docs
from cellrank.tl._utils import _maybe_subset_hvgs
from cellrank.external.kernels._utils import MarkerGenes
from cellrank.tl.kernels._exp_time_kernel import LastTimePoint

import numpy as np
import pandas as pd

_error = None
try:
    import wot

    from cellrank.tl.kernels import TransportMapKernel as Kernel
except ImportError as e:
    from cellrank.external.kernels._import_error_kernel import ErroredKernel as Kernel

    _error = e
    wot = None


class nstr(str):  # used for params+cache (precomputed cost matrices)
    """String class that is not equal to any other string."""

    def __eq__(self, other: str) -> bool:
        return False


@d.dedent
class WOTKernel(Kernel, error=_error):
    """
    Waddington optimal transport kernel from :cite:`schiebinger:19`.

    This class requires the `wot` package, which can be installed as
    `pip install git+https://github.com/broadinstitute/wot`.

    Parameters
    ----------
    %(adata)s
    %(backward)s
    time_key
        Key in :attr:`adata` ``.obs`` where experimental time is stored.
        The experimental time can be of either of a numeric or an ordered categorical type.
    %(cond_num)s

    Examples
    --------
    Workflow::

        # import packages, load data
        import scanpy as sc
        import cellrank as cr
        adata = cr.datasets.lung()

        # filter, normalize and annotate highly variable genes
        sc.pp.filter_genes(adata, min_cells=10)
        sc.pp.normalize_total(adata)
        sc.pp.log1p(adata)
        sc.pp.highly_variable_genes(adata)

        # estimate proliferation and apoptosis from gene sets (see. e.g. WOT tutorial for example lists)
        proliferation_genes = ...
        apoptosis_genes = ...
        sc.tl.score_genes(adata, gene_list=proliferation_genes, score_name='proliferation')
        sc.tl.score_genes(adata, gene_list=apoptosis_genes, score_name='apoptosis')

        # initialize kernel, estimate initial growth rate based on scores from above
        from cellrank.external.kernels import WOTKernel
        ot = WOTKernel(adata, time_key='day')
        ot.compute_initial_growth_rates(proliferation_key='proliferation',
                                        apoptosis_key='apoptosis',
                                        key_added='initial_growth_rates')

        # compute transport maps, aggregate into one single transition matrix
        ot.compute_transition_matrix(growth_rate_key='initial_growth_rates', growth_iters=3)
    """

    __import_error_message__ = (
        "Unable to import the kernel. Please install `wot` first as "
        "`pip install git+https://github.com/broadinstitute/wot`."
    )

    def __init__(
        self,
        adata: AnnData,
        backward: bool = False,
        time_key: str = "exp_time",
        compute_cond_num: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            adata,
            backward=backward,
            time_key=time_key,
            compute_cond_num=compute_cond_num,
            **kwargs,
        )

        # WOT's requirements
        cats = self.experimental_time.cat.categories
        self._exp_time = self.experimental_time.cat.rename_categories(
            dict(zip(cats, map(float, cats)))
        )
        self.adata.obs[self._time_key] = self.experimental_time
        self._growth_rates = None

    def _read_from_adata(
        self,
        conn_key: Optional[str] = "connectivities",
        read_conn: bool = True,
        **kwargs: Any,
    ) -> None:
        super()._read_from_adata(conn_key=conn_key, read_conn=False, **kwargs)

    def compute_initial_growth_rates(
        self,
        proliferation_key: Optional[str] = None,
        apoptosis_key: Optional[str] = None,
        organism: Optional[Literal["human", "mouse"]] = None,
        beta_min: float = 0.3,
        beta_max: float = 1.7,
        delta_min: float = 0.3,
        delta_max: float = 1.7,
        key_added: Optional[str] = None,
        **kwargs: Any,
    ) -> Optional[pd.Series]:
        r"""
        Estimate initial growth rates using a birth-death process as described in :cite:`schiebinger:19`.

        The doubling time is defined as :math:`\frac{\ln 2}{\beta - \delta}` (similarly defined for half-time).
        The logistic function is used to transform the birth/death rate scores and to smoothly interpolate between
        specified minimum and maximum birth/death rates.

        Parameters
        ----------
        proliferation_key
            Key in :attr:`anndata.AnnData.obs` where the birth rate score is saved.
        apoptosis_key
            Key in :attr:`anndata.AnnData.obs` where the death rate score is saved.
        organism
            Organism for which to calculate the birth/death scores, if they cannot be found in :attr:`adata`.
            In this case, :func:`scanpy.tl.score_genes` is used to calculate the scores based on an organism-dependent
            set of marker genes for proliferation and apoptosis.
        beta_min
            Minimum birth rate.
        beta_max
            Maximum birth rate.
        delta_min
            Minimum death rate.
        delta_max
            Maximum death rate.
        key_added
            Key in :attr:`adata` ``.obs`` where to add the estimated growth rates. If `None`, just return them.
        kwargs
            Keyword arguments for :func:`scanpy.tl.score_genes`. Only used when ``proliferation_key``
            or ``apoptosis_key`` cannot be found in :attr:`adata.AnnData.obs`.

        Returns
        -------
        :class:`pandas.Series`
            The estimated initial growth rates if ``key_added = None``, otherwise `None`.

        Notes
        -----
        If you don't have access to proliferation/apoptosis gene sets, you can use the ones defined in :mod:`cellrank`
        for a specific organism. Alternatively, you can also use WOT without an estimate of initial growth rates. In
        that case, make sure to use several iterations in
        :meth:`cellrank.external.kernels.WOTKernel.compute_transition_matrix` by increasing the ``growth_iters``
        parameter. A value around 3 works well in most cases.

        The markers used here were taken from the following sources:

            - human, proliferation - :cite:`tirosh:16:science`.
            - human, apoptosis - `Hallmark Apoptosis, MSigDB <https://www.gsea-msigdb.org/gsea/msigdb/cards/HALLMARK_APOPTOSIS>`_.
            - mouse, proliferation - :cite:`tirosh:16:nature`.
            - mouse, apoptosis - `Hallmark P53 Pathway, MSigDB <https://www.gsea-msigdb.org/gsea/msigdb/cards/HALLMARK_P53_PATHWAY>`_.

        For more information about WOT, see the official `tutorial <https://broadinstitute.github.io/wot/tutorial/>`_.
        """  # noqa: E501

        def get_scores(
            key: str,
            *,
            kind: Literal["proliferation", "apoptosis"],
            organism: Optional[Literal["human", "mouse"]],
        ) -> np.ndarray:
            try:
                return np.asarray(self.adata.obs[key])
            except KeyError:
                if organism is None:
                    raise KeyError(
                        f"Unable to find `{kind}` scores in `adata.obs[{kind!r}]`. Consider specifying `organism=...`."
                    ) from None

                logg.info(f"Computing `{kind}` scores")
                score_name = f"{kind}_score" if key is None else key

                sc.tl.score_genes(
                    self.adata,
                    gene_list=getattr(MarkerGenes, f"{kind}_markers")(organism),
                    score_name=score_name,
                    **kwargs,
                )

                return get_scores(score_name, kind=kind, organism=None)

        def logistic(x: np.ndarray, L: float, k: float, x0: float = 0) -> np.ndarray:
            return L / (1 + np.exp(-k * (x - x0)))

        def gen_logistic(
            p: np.ndarray, beta_max: float, beta_min: float, center: float, width: float
        ) -> np.ndarray:
            return beta_min + logistic(p, L=beta_max - beta_min, k=4 / width, x0=center)

        def beta(
            p: np.ndarray,
            beta_max: float = 1.7,
            beta_min: float = 0.3,
            center: float = 0.25,
        ) -> np.ndarray:
            return gen_logistic(p, beta_max, beta_min, center, width=0.5)

        def delta(
            a: np.ndarray,
            delta_max: float = 1.7,
            delta_min: float = 0.3,
            center: float = 0.1,
        ) -> np.ndarray:
            return gen_logistic(a, delta_max, delta_min, center, width=0.2)

        birth = beta(
            get_scores(proliferation_key, kind="proliferation", organism=organism),
            beta_min=beta_min,
            beta_max=beta_max,
        )
        death = delta(
            get_scores(apoptosis_key, kind="apoptosis", organism=organism),
            delta_min=delta_min,
            delta_max=delta_max,
        )
        gr = np.exp(birth - death)

        if key_added is None:
            return pd.Series(gr, index=self.adata.obs_names)
        self.adata.obs[key_added] = gr

    @inject_docs(ltp=LastTimePoint)
    def compute_transition_matrix(
        self,
        cost_matrices: Optional[
            Union[str, Mapping[Tuple[float, float], np.ndarray]]
        ] = None,
        lambda1: float = 1,
        lambda2: float = 50,
        epsilon: float = 0.05,
        growth_iters: int = 1,
        solver: Literal["fixed_iters", "duality_gap"] = "duality_gap",
        growth_rate_key: Optional[str] = None,
        use_highly_variable: Optional[Union[str, bool]] = True,
        last_time_point: Literal[
            "uniform", "diagonal", "connectivities"
        ] = LastTimePoint.UNIFORM,
        threshold: Optional[Union[float, Literal["auto"]]] = "auto",
        conn_kwargs: Mapping[str, Any] = MappingProxyType({}),
        **kwargs: Any,
    ) -> "WOTKernel":
        """
        Compute transition matrix using Waddington OT :cite:`schiebinger:19`.

        Computes transport maps linking together pairs of time points for time-series single cell data using unbalanced
        optimal transport, taking into account cell birth and death rates. From the sequence of transition maps linking
        pairs of sequential time points, we construct one large transition matrix which contains the normalized
        transport maps as blocks on the 1st upper diagonal.

        Parameters
        ----------
        cost_matrices
            Cost matrices for each consecutive pair of time points.
            If a :class:`str`, it specifies a key in :attr:`anndata.AnnData.layers` or :attr:`anndata.AnnData.obsm`.
            containing cell features that are used to compute cost matrices. If `None`, use `WOT`'s default, i.e.
            compute distances in PCA space derived from :attr:`anndata.AnnData.X` for each time point pair separately.
        lambda1
            Regularization parameter for the marginal constraint on :math:`p`, the transport map row sums.
            Smaller value is useful when precise information about the growth rate is not present.
        lambda2
            Regularization parameter for the marginal constraint on :math:`q`, the transport map column sums.
        epsilon
            Entropy regularization parameter. Larger value gives more entropic descendant distributions.
        growth_iters
            Number of iterations for growth rate estimates. If growth rates are not known, consider using more
            iterations.
        solver
            Which solver to use.
        growth_rate_key
            Key in :attr:`anndata.AnnData.obs` where initial cell growth rates are stored.
            See :meth:`compute_initial_growth_rates` on how to estimate them.
        use_highly_variable
            Key in :attr:`anndata.AnnData.var` where highly variable genes are stored.
            If `True`, use `'highly_variable'`. If `None`, use all genes.
        last_time_point
            How to define transitions within the last time point. Valid options are:

                - `{ltp.UNIFORM!r}` - row-normalized matrix of 1s for transitions within the last time point.
                - `{ltp.DIAGONAL!r}` - diagonal matrix with 1s on the diagonal.
                - `{ltp.CONNECTIVITIES!r}` - use transitions from :class:`cellrank.tl.kernels.ConnectivityKernel`
                  derived from the last time point subset of :attr:`adata`.
        threshold
            How to remove small non-zero values from the transition matrix. Valid options are:

                - `'auto'` - find the maximum threshold value which will not remove every non-zero value from any row.
                - :class:`float` - value in `[0, 100]` corresponding to a percentage of non-zeros to remove.
                  Rows where all values are removed will have uniform distribution.
                - `None` - do not threshold.

        conn_kwargs
            Keyword arguments for :func:`scanpy.pp.neighbors`, when using ``last_time_point = {ltp.CONNECTIVITIES!r}``.
            Can contain `'density_normalize'` for
            :meth:`cellrank.tl.kernels.ConnectivityKernel.compute_transition_matrix`.
        kwargs
            Additional keyword arguments for optimal transport configuration.

        Returns
        -------
        :class:`cellrank.external.kernels.WOTKernel`
            Makes :attr:`transition_matrix`, :attr:`transport_maps` and :attr:`growth_rates` available.
            Also modifies :attr:`anndata.AnnData.obs` with the following key:

                - `'estimated_growth_rates'` - the estimated final growth rates.

        Notes
        -----
        For more information about WOT, see the official `tutorial <https://broadinstitute.github.io/wot/tutorial/>`_.
        """
        # disallow these params (because they e.g. subset the data)
        _ = kwargs.pop("cell_day_filter", None)
        _ = kwargs.pop("covariate_field", None)
        _ = kwargs.pop("ncounts", None)
        _ = kwargs.pop("ncells", None)
        _ = kwargs.pop("parameters", None)  # parameters file
        kwargs["lambda1"] = lambda1
        kwargs["lambda2"] = lambda2
        kwargs["epsilon"] = epsilon
        kwargs["growth_iters"] = max(growth_iters, 1)
        last_time_point = LastTimePoint(last_time_point)

        start = logg.info(
            "Computing transition matrix using Waddington optimal transport"
        )

        adata = _maybe_subset_hvgs(self.adata, use_highly_variable=use_highly_variable)
        cost_matrices, cmat_param = self._generate_cost_matrices(adata, cost_matrices)
        if self._reuse_cache(
            {
                "cost_matrices": cmat_param,
                "solver": solver,
                "growth_rate_key": growth_rate_key,
                "use_highly_variable": use_highly_variable,
                "last_time_point": last_time_point,
                "threshold": threshold,
                **kwargs,
            },
            time=start,
        ):
            return self

        tmap = self._compute_pairwise_tmaps(
            adata,
            cost_matrices=cost_matrices,
            solver=solver,
            growth_rate_field=growth_rate_key,
            **kwargs,
        )
        tmap = self._restich_tmaps(tmap, last_time_point, conn_kwargs=conn_kwargs)
        self._growth_rates = tmap.obs

        self._compute_transition_matrix(
            matrix=tmap.X,
            density_normalize=False,
            check_irreducibility=False,
        )
        if threshold:
            self._threshold_transition_matrix(threshold)
        self.adata.obs["estimated_growth_rates"] = self.growth_rates[f"g{growth_iters}"]

        logg.info("    Finish", time=start)

        return self

    def _compute_pairwise_tmaps(
        self,
        adata: AnnData,
        cost_matrices: Optional[
            Union[str, Mapping[Tuple[float, float], np.ndarray]]
        ] = None,
        solver: Literal["fixed_iters", "duality_gap"] = "duality_gap",
        growth_rate_field: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[Tuple[float, float], AnnData]:
        _ = wot.ot.OTModel(
            adata,
            day_field=self._time_key,
            covariate_field=None,
            growth_rate_field=growth_rate_field,
        )
        for k in kwargs:
            if k not in _.ot_config:
                raise TypeError(f"WOT got an unexpected keyword argument {k!r}.")

        self._ot_model = wot.ot.OTModel(
            adata,
            day_field=self._time_key,
            covariate_field=None,
            growth_rate_field=growth_rate_field,
            solver=solver,
            **kwargs,
        )

        self._tmaps: Dict[Tuple[float, float], AnnData] = {}
        start = logg.info(
            f"Computing transport maps for `{len(cost_matrices)}` time pairs"
        )
        for tpair, cost_matrix in tqdm(cost_matrices.items(), unit="time pair"):
            tmap: Optional[AnnData] = self._ot_model.compute_transport_map(
                *tpair, cost_matrix=cost_matrix
            )
            if tmap is None:
                raise TypeError(
                    f"Unable to compute transport map for time pair `{tpair}`. "
                    f"Please ensure `adata.obs[{self._time_key!r}]` has the correct dtype (float)."
                )
            tmap.X = tmap.X.astype(np.float64)
            nans = int(np.sum(~np.isfinite(tmap.X)))
            if nans:
                raise ValueError(
                    f"Encountered `{nans}` non-finite values for time pair `{tpair}`."
                )
            self._tmaps[tpair] = tmap

        logg.info("    Finish", time=start)

        return self._tmaps

    def _generate_cost_matrices(
        self,
        adata: AnnData,
        cost_matrices: Optional[
            Union[str, Mapping[Tuple[float, float], np.ndarray]]
        ] = None,
    ) -> Tuple[Mapping[Tuple[float, float], Optional[np.ndarray]], str]:
        timepoints = self.experimental_time.cat.categories
        timepoints = list(zip(timepoints[:-1], timepoints[1:]))

        if cost_matrices is None:
            logg.info("Using default cost matrices")
            return {tpair: None for tpair in timepoints}, "default"

        if isinstance(cost_matrices, dict):
            logg.info("Using precomputed cost matrices")

            cmats = {}
            for tpair in timepoints:
                if tpair not in cost_matrices:
                    logg.warning(
                        f"Unable to find cost matrix for pair `{tpair}`. Using default"
                    )
                cmats[tpair] = cmat = cost_matrices.get(tpair, None)
                if cmat is not None:
                    n_start = len(np.where(self.experimental_time == tpair[0])[0])
                    n_end = len(np.where(self.experimental_time == tpair[1])[0])
                    try:
                        if cmat.shape != (n_start, n_end):
                            raise ValueError(
                                f"Expected cost matrix for time pair `{tpair}` to be "
                                f"of shape `{(n_start, n_end)}`, found `{cmat.shape}`."
                            )
                    except AttributeError:
                        logg.warning(
                            f"Unable to verify whether supplied cost matrix for time pair `{tpair}` "
                            f"has the correct shape `{(n_start, n_end)}`"
                        )

            # prevent equality comparison when comparing with cache
            return cmats, nstr("precomputed")

        if isinstance(cost_matrices, str):
            logg.info(f"Computing cost matrices using `{cost_matrices!r}` key")
            if cost_matrices == "X":
                cost_matrices = None

            try:
                features = adata._get_X(layer=cost_matrices)
                modifier = "layer"
            except KeyError:
                try:
                    features = adata.obsm[cost_matrices]
                    modifier = "obsm"
                except KeyError:
                    raise KeyError(
                        f"Unable to find key `{cost_matrices!r}` in `adata.layers` or `adata.obsm`."
                    ) from None

            cmats = {}
            for tpair in tqdm(timepoints, unit="cost matrix"):
                start_ixs = np.where(self.experimental_time == tpair[0])[0]
                end_ixs = np.where(self.experimental_time == tpair[1])[0]

                # being sparse is handled in WOT's function below
                cmats[tpair] = wot.ot.OTModel.compute_default_cost_matrix(
                    features[start_ixs], features[end_ixs]
                )

            return cmats, f"{modifier}:{cost_matrices}"

        raise NotImplementedError(
            f"Specifying cost matrices as "
            f"`{type(cost_matrices).__name__}` is not yet implemented."
        )

    def _threshold_transition_matrix(
        self, threshold: Union[float, Literal["auto"]]
    ) -> None:
        tmat = self.transition_matrix
        if threshold == "auto":
            threshold = min(np.max(tmat[i].data) for i in range(tmat.shape[0]))
            logg.info(f"Using `threshold={threshold}`")
            tmat.data[tmat.data < threshold] = 0.0
        else:
            if not (0 <= threshold <= 100):
                raise ValueError(
                    f"Expected `threshold to be in `[0, 100]`, found `{threshold}`.`"
                )
            threshold = np.percentile(tmat.data, threshold)
            logg.info(f"Using `threshold={threshold}`")
            tmat.data[tmat.data <= threshold] = 0.0

        tmat.eliminate_zeros()

        self._compute_transition_matrix(
            matrix=tmat,
            density_normalize=False,
            check_irreducibility=False,
        )

    @property
    def growth_rates(self) -> Optional[pd.DataFrame]:
        """Estimated cell growth rates for each growth rate iteration."""
        return self._growth_rates

    def __invert__(self) -> "WOTKernel":
        super().__invert__()
        # because WOT reads from `adata`
        self.adata.obs[self._time_key] = self.experimental_time
        return self



from typing import List, Tuple, Union, Optional, Sequence

import os
import pytest
from copy import deepcopy
from enum import Enum
from _helpers import assert_array_nan_equal, assert_estimators_equal

import cellrank as cr
from anndata import AnnData
from cellrank.tl import Lineage
from cellrank._key import Key
from cellrank.tl.kernels import VelocityKernel, ConnectivityKernel

import numpy as np
import pandas as pd
from scipy.sparse import issparse
from pandas.testing import assert_frame_equal, assert_series_equal


# fmt: off
class State(str, Enum):
    SCHUR = "schur"
    MACRO = "macro"
    TERM = "term"
    ABS = "abs"
    LIN = "lin"

    @property
    def prev(self) -> Optional["State"]:
        if self.value == "schur":
            return None
        if self.value == "macro":
            return State.SCHUR
        if self.value == "term":
            return State.MACRO
        if self.value == "abs":
            return State.TERM
        if self.value == "lin":
            return State.ABS
        return None

    @property
    def next(self) -> Optional["State"]:
        if self.value == "schur":
            return State.MACRO
        if self.value == "macro":
            return State.TERM
        if self.value == "term":
            return State.ABS
        if self.value == "abs":
            return State.LIN
        if self.value == "lin":
            return None
        return None

    @property
    def key(self) -> Optional[str]:
        bwd = False
        if self.value == "schur":
            return f"schur_decomposition_{Key.backward(bwd)}"
        if self.value == "macro":
            return Key.obs.macrostates(bwd)
        if self.value == "term":
            return Key.obs.term_states(bwd)
        if self.value == "abs":
            return Key.obsm.abs_probs(bwd)
        if self.value == "lin":
            return Key.varm.lineage_drivers(bwd)
        return None

    @property
    def attr_keys(self) -> Optional[Sequence[Tuple[str, str]]]:
        bwd = False
        if self.value == "schur":
            key1 = Key.uns.eigen(bwd)
            key2 = Key.obsm.schur_vectors(bwd)
            key3 = Key.uns.schur_matrix(bwd)
            return ("uns", key1), ("obsm", key2), ("uns", key3)
        if self.value == "macro":
            key1 = Key.obs.macrostates(bwd)
            key2 = Key.uns.colors(key1)
            key3 = Key.obsm.memberships(key1)
            return ("obs", key1), ("uns", key2), ("obsm", key3)
        if self.value == "term":
            key1 = Key.obs.term_states(bwd)
            key2 = Key.obs.probs(key1)
            key3 = Key.uns.colors(key1)
            return ("obs", key1), ("obs", key2), ("uns", key3)
        if self.value == "abs":
            key1 = Key.obsm.abs_probs(bwd)
            key2 = Key.obsm.abs_times(bwd)
            key3 = Key.obs.priming_degree(bwd)
            return ("obsm", key1), ("obsm", key2), ("obs", key3)
        if self.value == "lin":
            key1 = Key.varm.lineage_drivers(bwd)
            return ("varm", key1),
        return None

    @property
    def attrs(self) -> Optional[Sequence[Tuple[str, type]]]:
        if self.value == "schur":
            return ("_eigendecomposition", dict), ("_schur_vectors", np.ndarray), ("_schur_matrix", np.ndarray)
        if self.value == "macro":
            return (
                ("_macrostates", pd.Series), ("_macrostates_colors", np.ndarray),
                ("_macrostates_memberships", Lineage), ("_coarse_tmat", pd.DataFrame),
                ("_coarse_init_dist", pd.Series), ("_coarse_stat_dist", pd.Series)
            )
        if self.value == "term":
            return ("_term_states", pd.Series), ("_term_states_probs", pd.Series), ("_term_states_colors", np.ndarray)
        if self.value == "abs":
            return (
                ("_absorption_probabilities", Lineage), ("_absorption_times", pd.DataFrame),
                ("_priming_degree", pd.Series)
            )
        if self.value == "lin":
            return ("_lineage_drivers", pd.DataFrame),
        return None

# fmt: on


def shares_mem(x, y) -> bool:
    if isinstance(x, Lineage):
        return np.shares_memory(x.X, y.X)
    if issparse(x):
        return np.shares_memory(x.data, y.data)
    return np.shares_memory(x, y)


def _check_eigdecomposition(mc: cr.tl.estimators.GPCCA) -> None:
    assert isinstance(mc.eigendecomposition, dict)
    assert set(mc.eigendecomposition.keys()) == {
        "D",
        "eigengap",
        "params",
    }
    assert "V_l" not in mc.eigendecomposition
    assert "V_r" not in mc.eigendecomposition
    assert "stationary_dist" not in mc.eigendecomposition

    assert Key.uns.eigen(mc.backward) in mc.adata.uns


def _check_compute_macro(mc: cr.tl.estimators.GPCCA) -> None:
    assert isinstance(mc.macrostates, pd.Series)
    assert len(mc._macrostates_colors) == len(mc.macrostates.cat.categories)

    if "stationary_dist" in mc.eigendecomposition:  # one state
        assert isinstance(mc.macrostates_memberships, cr.tl.Lineage)
        assert mc.macrostates_memberships.shape[1] == 1
        np.testing.assert_allclose(mc.macrostates_memberships.X.sum(), 1.0)

        assert mc.schur_matrix is None
        assert mc.schur_vectors is None
        assert mc.coarse_initial_distribution is None
        assert mc.coarse_stationary_distribution is None
        assert mc.coarse_T is None
    else:
        assert isinstance(mc.macrostates_memberships, cr.tl.Lineage)
        if mc.macrostates_memberships.shape[1] > 1:
            np.testing.assert_allclose(mc.macrostates_memberships.sum(1), 1.0)

        assert isinstance(mc.schur_matrix, np.ndarray)
        assert isinstance(mc.schur_vectors, np.ndarray)
        assert isinstance(mc.coarse_initial_distribution, pd.Series)
        assert isinstance(mc.coarse_T, pd.DataFrame)
        np.testing.assert_array_equal(mc.coarse_T.index, mc.coarse_T.columns)
        np.testing.assert_array_equal(
            mc.coarse_T.index, mc.coarse_stationary_distribution.index
        )
        if mc.coarse_stationary_distribution is not None:
            assert isinstance(mc.coarse_stationary_distribution, pd.Series)
            np.testing.assert_array_equal(
                mc.coarse_T.index, mc.coarse_stationary_distribution.index
            )


def _check_renaming_no_write_terminal(mc: cr.tl.estimators.GPCCA) -> None:
    assert mc.terminal_states is None
    assert mc.terminal_states_probabilities is None
    assert mc.terminal_states_memberships is None

    key = Key.obs.term_states(mc.backward)
    assert key not in mc.adata.obs
    assert Key.obs.probs(key) not in mc.adata.obs
    assert Key.uns.colors(key) not in mc.adata.uns


def _check_abs_probs(mc: cr.tl.estimators.GPCCA) -> None:
    # fmt: off
    # macrostates
    assert isinstance(mc.macrostates, pd.Series)
    assert isinstance(mc.macrostates_memberships, cr.tl.Lineage)
    np.testing.assert_array_equal(mc._macrostates_colors, mc.macrostates_memberships.colors)

    # term states
    key = Key.obs.term_states(mc.backward)
    assert isinstance(mc.terminal_states, pd.Series)
    # TODO(michalk8): assert series equal from pandas
    assert_array_nan_equal(mc.adata.obs[key], mc.terminal_states)
    np.testing.assert_array_equal(mc.adata.uns[Key.uns.colors(key)], mc.absorption_probabilities.colors)
    np.testing.assert_array_equal(mc.adata.uns[Key.uns.colors(key)], mc._term_states_colors)
    assert isinstance(mc.terminal_states_probabilities, pd.Series)
    np.testing.assert_array_equal(mc.adata.obs[Key.obs.probs(key)], mc.terminal_states_probabilities)

    # abs probs
    key = Key.obsm.abs_probs(mc.backward)
    assert isinstance(mc.absorption_probabilities, cr.tl.Lineage)
    np.testing.assert_array_almost_equal(mc.absorption_probabilities.sum(1), 1.0)
    assert isinstance(mc.adata.obsm[key], cr.tl.Lineage)
    np.testing.assert_array_equal(mc.adata.obsm[key], mc.absorption_probabilities.X)

    # priming
    key = Key.obs.priming_degree(mc.backward)
    assert isinstance(mc.priming_degree, pd.Series)
    np.testing.assert_array_equal(mc.adata.obs[key], mc.priming_degree)
    # fmt: on


def _fit_gpcca(adata, state: str, backward: bool = False) -> cr.estimators.GPCCA:
    vk = VelocityKernel(adata, backward=backward).compute_transition_matrix(
        softmax_scale=4
    )
    ck = ConnectivityKernel(adata, backward=backward).compute_transition_matrix()
    terminal_kernel = 0.8 * vk + 0.2 * ck

    mc = cr.tl.estimators.GPCCA(terminal_kernel)
    mc.compute_eigendecomposition()
    mc.compute_schur(n_components=10, method="krylov")
    if state == State.SCHUR:
        return mc
    mc.compute_macrostates(n_states=2)
    if state == State.MACRO:
        return mc
    mc.set_terminal_states_from_macrostates()
    if state == State.TERM:
        return mc
    mc.compute_absorption_probabilities(time_to_absorption="all")
    mc.compute_lineage_priming()
    if state == State.ABS:
        return mc
    mc.compute_lineage_drivers(use_raw=False, cluster_key="clusters")
    if state == State.LIN:
        return mc

    raise NotImplementedError(state)


def _assert_params(
    g: cr.tl.estimators.GPCCA,
    state: Optional[State],
    fwd: bool = False,
    init: bool = True,
) -> None:
    if state is None:
        return

    key = state.key
    if fwd:
        if init:
            assert isinstance(g.params[key], dict), state
        else:
            assert g.params.get(key, {}) == {}, state
        _assert_params(g, state.next, fwd=True, init=False)
    else:
        assert isinstance(g.params[key], dict), state
        _assert_params(g, state.prev, fwd=False)


def _assert_adata(
    adata: AnnData, state: Optional[State], fwd: bool = False, init: bool = True
) -> None:
    if state is None:
        return

    if fwd:
        if init:
            for attr, key in state.attr_keys:
                obj = getattr(adata, attr)
                assert key in obj, sorted(obj.keys())
        else:
            for attr, key in state.attr_keys:
                obj = getattr(adata, attr)
                assert key not in obj, sorted(obj.keys())
        _assert_adata(adata, state.next, fwd=True, init=False)
    else:
        for attr, key in state.attr_keys:
            obj = getattr(adata, attr)
            assert key in obj, sorted(obj.keys())
        _assert_adata(adata, state.prev, fwd=False)


def _assert_gpcca_attrs(
    g: cr.tl.estimators.GPCCA,
    state: Optional[State] = None,
    fwd: bool = False,
    init: bool = True,
) -> None:
    if state is None:
        return
    if fwd:
        if init:
            for attr, dtype in state.attrs:
                obj = getattr(g, attr)
                assert isinstance(obj, dtype)
        else:
            for attr, dtype in state.attrs:
                obj = getattr(g, attr)
                assert obj is None, attr
        _assert_gpcca_attrs(g, state.next, fwd=True, init=False)
    else:
        for attr, dtype in state.attrs:
            obj = getattr(g, attr)
            assert isinstance(obj, dtype), attr
        _assert_gpcca_attrs(g, state.prev, fwd=False)


class TestGPCCA:
    def test_compute_eigendecomposition(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_eigendecomposition(k=2, only_evals=True)

        _check_eigdecomposition(mc)

    def test_compute_schur_invalid_n_comps(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=1, method="krylov")

        assert mc.schur_vectors.shape[1] == 2

    def test_compute_schur_invalid_method(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        with pytest.raises(ValueError):
            mc.compute_schur(method="foobar")

    def test_compute_schur_invalid_eig_sort(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        with pytest.raises(ValueError):
            mc.compute_schur(which="foobar", method="krylov")

    def test_compute_schur_write_eigvals(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        _check_eigdecomposition(mc)

    def test_compute_schur_write_eigvals_similar_to_orig_eigdecomp(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_eigendecomposition(k=10, only_evals=True)

        _check_eigdecomposition(mc)
        orig_ed = deepcopy(mc.eigendecomposition)

        mc._eigendecomposition = None
        mc.compute_schur(n_components=10, method="krylov")

        _check_eigdecomposition(mc)
        schur_ed = mc.eigendecomposition

        assert orig_ed.keys() == schur_ed.keys()
        assert orig_ed["eigengap"] == schur_ed["eigengap"]
        n = min(orig_ed["params"]["k"], schur_ed["params"]["k"])
        np.testing.assert_array_almost_equal(
            orig_ed["D"].real[:n], schur_ed["D"].real[:n]
        )
        np.testing.assert_array_almost_equal(
            np.abs(orig_ed["D"].imag[:n]), np.abs(schur_ed["D"].imag[:n])
        )  # complex conj.

    def test_compute_macrostates_no_eig(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        with pytest.raises(RuntimeError):
            mc.compute_macrostates(n_states=None)

    def test_compute_macrostates_1_state_no_eig(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_macrostates(n_states=1)

    def test_compute_macro_none_states(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_eigendecomposition(only_evals=True)
        mc.compute_macrostates(n_states=None)

        _check_compute_macro(mc)

    def test_compute_macrostates_schur(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=2)

        _check_compute_macro(mc)

    def test_compute_macrostates_min_chi_too_low_min(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=[1, 1])

        _check_compute_macro(mc)

    def test_compute_macrostates_min_chi_inverted_range(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=[4, 2])

        _check_compute_macro(mc)

    def test_compute_macrostates_min_chi_range_same_values(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=[2, 2])

        _check_compute_macro(mc)

    def test_compute_macrostates_min_chi_normal_run(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=[2, 4])

        _check_compute_macro(mc)

    def test_compute_macro_invalid_cluster_key(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        with pytest.raises(KeyError):
            mc.compute_macrostates(n_states=2, cluster_key="foobar")

    def test_compute_macro_cache(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=11, method="krylov")

        mc.compute_macrostates(n_states=2)

        assert mc.schur_vectors.shape[1] == 11
        np.testing.assert_array_equal(mc.schur_matrix.shape, [11, 11])

    def test_set_initial_states_from_forward(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large, backward=False).compute_transition_matrix(
            softmax_scale=4
        )
        ck = ConnectivityKernel(adata_large, backward=False).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2, n_cells=5)
        obsm_keys = set(mc.adata.obsm.keys())
        # TODO
        mc._set_initial_states_from_macrostates("0")

        key = Key.obs.term_states(not mc.backward)

        assert key in mc.adata.obs
        np.testing.assert_array_equal(mc.adata.obs[key].cat.categories, ["0"])
        assert Key.obs.probs(key) in mc.adata.obs
        assert Key.uns.colors(key) in mc.adata.uns

        # make sure that we don't write anything there - it's useless
        assert set(mc.adata.obsm.keys()) == obsm_keys

    def test_compute_initial_states_from_forward_no_macro(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large, backward=False).compute_transition_matrix(
            softmax_scale=4
        )
        ck = ConnectivityKernel(adata_large, backward=False).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        with pytest.raises(RuntimeError):
            # TODO
            mc._compute_initial_states(1)

    def test_compute_initial_states_from_forward_too_many_states(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large, backward=False).compute_transition_matrix(
            softmax_scale=4
        )
        ck = ConnectivityKernel(adata_large, backward=False).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2, n_cells=5)
        with pytest.raises(ValueError):
            mc._compute_initial_states(42)

    def test_compute_initial_states_from_forward_too_few_states(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large, backward=False).compute_transition_matrix(
            softmax_scale=4
        )
        ck = ConnectivityKernel(adata_large, backward=False).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2, n_cells=5)
        with pytest.raises(ValueError):
            mc._compute_initial_states(0)

    def test_compute_initial_states_from_forward_no_stat_dist(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large, backward=False).compute_transition_matrix(
            softmax_scale=4
        )
        ck = ConnectivityKernel(adata_large, backward=False).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2, n_cells=5)
        mc._coarse_stat_dist = None

        with pytest.raises(ValueError):
            mc._compute_initial_states(0)

    def test_compute_initial_states_from_forward_normal_run(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large, backward=False).compute_transition_matrix(
            softmax_scale=4
        )
        ck = ConnectivityKernel(adata_large, backward=False).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2, n_cells=5)
        obsm_keys = set(mc.adata.obsm.keys())
        csd = mc.coarse_stationary_distribution
        expected = csd.index[np.argmin(csd)]

        mc._compute_initial_states(1)

        key = Key.obs.term_states(not mc.backward)

        assert key in mc.adata.obs
        np.testing.assert_array_equal(mc.adata.obs[key].cat.categories, [expected])
        assert Key.obs.probs(key) in mc.adata.obs
        assert Key.uns.colors(key) in mc.adata.uns

        # make sure that we don't write anything there - it's useless
        assert set(mc.adata.obsm.keys()) == obsm_keys

    def test_set_terminal_states_from_macrostates(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2, n_cells=5)
        mc.set_terminal_states_from_macrostates()
        mc.compute_absorption_probabilities()
        mc.compute_lineage_priming()

        _check_abs_probs(mc)

    def test_set_terminal_states_from_macrostates_rename_states_invalid(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2, n_cells=5)
        with pytest.raises(ValueError):
            mc.set_terminal_states_from_macrostates({"0": "1"})
        _check_renaming_no_write_terminal(mc)

    def test_set_terminal_states_from_macrostates_rename_not_unique_new_names(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=3, n_cells=5)
        with pytest.raises(ValueError):
            mc.set_terminal_states_from_macrostates({"0": "1", "1": "1"})
        _check_renaming_no_write_terminal(mc)

    def test_set_terminal_states_from_macrostates_rename_overlapping_old_keys(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=3, n_cells=5)
        with pytest.raises(ValueError):
            mc.set_terminal_states_from_macrostates({"0, 1": "foo", "1, 2": "bar"})
        _check_renaming_no_write_terminal(mc)

    def test_set_terminal_states_from_macrostates_rename_states(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2, n_cells=5)
        mc.set_terminal_states_from_macrostates({"0": "foo"})
        mc.compute_absorption_probabilities()
        mc.compute_lineage_priming()

        _check_abs_probs(mc)

        np.testing.assert_array_equal(mc.terminal_states.cat.categories, ["foo"])
        np.testing.assert_array_equal(mc.terminal_states_memberships.names, ["foo"])

    def test_set_terminal_states_from_macrostates_join_and_rename_states(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=3, n_cells=5)
        mc.set_terminal_states_from_macrostates({"0, 1": "foo", "2": "bar"})
        mc.compute_absorption_probabilities()
        mc.compute_lineage_priming()

        _check_abs_probs(mc)

        np.testing.assert_array_equal(mc.terminal_states.cat.categories, ["foo", "bar"])
        np.testing.assert_array_equal(
            mc.terminal_states_memberships.names, ["foo", "bar"]
        )

    def test_set_terminal_states_from_macrostates_no_cells(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2, n_cells=None)
        with pytest.raises(TypeError):
            mc.set_terminal_states_from_macrostates(n_cells=None)

    def test_set_terminal_states_from_macrostates_non_positive_cells(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2, n_cells=None)
        with pytest.raises(ValueError):
            mc.set_terminal_states_from_macrostates(n_cells=0)

    def test_set_terminal_states_from_macrostates_invalid_name(
        self, adata_large: AnnData
    ):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2)
        with pytest.raises(KeyError):
            mc.set_terminal_states_from_macrostates(names=["foobar"])

    @pytest.mark.parametrize("values", ["Astrocytes", ["Astrocytes", "OPC"]])
    def test_set_terminal_states_clusters(
        self, adata_large: AnnData, values: Union[str, List[str]]
    ):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        to_remove = list(
            set(adata_large.obs["clusters"].cat.categories)
            - ({values} if isinstance(values, str) else set(values))
        )
        expected = adata_large.obs["clusters"].cat.remove_categories(to_remove)

        mc = cr.tl.estimators.GPCCA(terminal_kernel)

        mc.set_terminal_states({"clusters": values})
        pd.testing.assert_series_equal(
            expected, mc.terminal_states, check_category_order=False, check_names=False
        )

    def test_compute_terminal_states_invalid_method(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2)
        with pytest.raises(ValueError):
            mc.compute_terminal_states(method="foobar")

    def test_compute_terminal_states_no_cells(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2)
        with pytest.raises(TypeError):
            mc.compute_terminal_states(n_cells=None)

    def test_compute_terminal_states_non_positive_cells(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2)
        with pytest.raises(ValueError):
            mc.compute_terminal_states(n_cells=0)

    def test_compute_terminal_states_eigengap(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2)
        mc.compute_terminal_states(n_cells=5, method="eigengap")
        mc.compute_absorption_probabilities()
        mc.compute_lineage_priming()

        _check_abs_probs(mc)

    def test_compute_terminal_states_n_main_states(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2)
        mc.compute_terminal_states(n_cells=5, method="top_n", n_states=1)
        mc.compute_absorption_probabilities()
        mc.compute_lineage_priming()

        _check_abs_probs(mc)

    def test_compute_terminal_states_stability(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck
        thresh = 0.5

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=5)
        mc.compute_terminal_states(
            n_cells=5, method="stability", stability_threshold=thresh
        )
        mc.compute_absorption_probabilities()
        mc.compute_lineage_priming()

        coarse_T = mc.coarse_T
        self_probs = pd.Series(np.diag(coarse_T), index=coarse_T.columns)
        names = self_probs[self_probs.values >= thresh].index

        np.testing.assert_array_equal(
            set(names), set(mc.terminal_states.cat.categories)
        )
        _check_abs_probs(mc)

    def test_compute_terminal_states_no_selected(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2)
        with pytest.raises(ValueError):
            mc.compute_terminal_states(
                n_cells=5, method="stability", stability_threshold=42
            )

    def test_compute_terminal_states_too_many_cells(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2)
        with pytest.raises(ValueError):
            mc.compute_terminal_states(n_cells=4200)

    def test_compute_terminal_states_default(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")

        mc.compute_macrostates(n_states=2)
        mc.compute_terminal_states(n_cells=5)
        mc.compute_absorption_probabilities()
        mc.compute_lineage_priming()

        _check_abs_probs(mc)

    def test_compute_lineage_drivers_invalid_lineages(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=2)
        mc.set_terminal_states_from_macrostates()
        mc.compute_absorption_probabilities()

        with pytest.raises(KeyError):
            mc.compute_lineage_drivers(use_raw=False, lineages=["foo"])

    def test_compute_lineage_drivers_invalid_clusters(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=2)
        mc.set_terminal_states_from_macrostates()
        mc.compute_absorption_probabilities()

        with pytest.raises(KeyError):
            mc.compute_lineage_drivers(
                use_raw=False, cluster_key="clusters", clusters=["foo"]
            )

    def test_compute_lineage_drivers_normal_run(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=2)
        mc.set_terminal_states_from_macrostates()
        mc.compute_absorption_probabilities()
        mc.compute_lineage_drivers(use_raw=False, cluster_key="clusters")

        key = Key.varm.lineage_drivers(False)
        for lineage in ["0", "1"]:
            assert np.all(mc.adata.varm[key][f"{lineage}_corr"] >= -1.0)
            assert np.all(mc.adata.varm[key][f"{lineage}_corr"] <= 1.0)

            assert np.all(mc.adata.varm[key][f"{lineage}_qval"] >= 0)
            assert np.all(mc.adata.varm[key][f"{lineage}_qval"] <= 1.0)

    def test_plot_lineage_drivers_not_computed(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=2)
        mc.set_terminal_states_from_macrostates()
        mc.compute_absorption_probabilities()

        with pytest.raises(RuntimeError):
            mc.plot_lineage_drivers("0")

    def test_plot_lineage_drivers_invalid_name(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=2)
        mc.set_terminal_states_from_macrostates()
        mc.compute_absorption_probabilities()
        mc.compute_lineage_drivers(use_raw=False, cluster_key="clusters")

        with pytest.raises(KeyError):
            mc.plot_lineage_drivers("foo", use_raw=False)

    def test_plot_lineage_drivers_invalid_n_genes(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=2)
        mc.set_terminal_states_from_macrostates()
        mc.compute_absorption_probabilities()
        mc.compute_lineage_drivers(use_raw=False, cluster_key="clusters")

        with pytest.raises(ValueError):
            mc.plot_lineage_drivers("0", use_raw=False, n_genes=0)

    def test_plot_lineage_drivers_normal_run(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=2)
        mc.set_terminal_states_from_macrostates()
        mc.compute_absorption_probabilities()
        mc.compute_lineage_drivers(use_raw=False, cluster_key="clusters")

        mc.plot_lineage_drivers("0", use_raw=False)

    def test_compute_priming_clusters(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=2)
        mc.set_terminal_states_from_macrostates()
        mc.compute_absorption_probabilities()

        cat = adata_large.obs["clusters"].cat.categories[0]
        deg1 = mc.compute_lineage_priming(
            method="kl_divergence", early_cells={"clusters": [cat]}
        )
        deg2 = mc.compute_lineage_priming(
            method="kl_divergence",
            early_cells=(adata_large.obs["clusters"] == cat).values,
        )

        assert_series_equal(deg1, deg2)
        # because passing it to a dataframe changes its name
        key = Key.obs.priming_degree(mc.backward)
        assert_series_equal(adata_large.obs[key], deg1, check_names=False)
        assert_series_equal(mc.priming_degree, deg1)

    def test_recompute_terminal_states_different_n_states(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        mc = cr.tl.estimators.GPCCA(terminal_kernel)
        mc.compute_schur(n_components=10, method="krylov")
        mc.compute_macrostates(n_states=2)
        mc.set_terminal_states_from_macrostates()

        key = Key.obsm.memberships(Key.obs.term_states(mc.backward))
        assert len(mc.terminal_states.cat.categories) == 2
        assert adata_large.obsm[key].shape == (adata_large.n_obs, 2)

        mc.set_terminal_states({"foo": adata_large.obs_names[:20]})

        assert len(mc.terminal_states.cat.categories) == 1
        assert mc.terminal_states_memberships is None
        assert key not in adata_large.obsm

    def test_fit_predict(self, adata_large: AnnData):
        vk = VelocityKernel(adata_large).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata_large).compute_transition_matrix()
        terminal_kernel = 0.8 * vk + 0.2 * ck

        g = cr.tl.estimators.GPCCA(terminal_kernel)
        tmp = g.fit(n_states=2)

        assert tmp is g
        _assert_params(g, State.MACRO, fwd=False)
        _assert_params(g, State.MACRO, fwd=True)

        res = g.predict()
        assert res is None
        _assert_params(g, State.TERM, fwd=False)
        _assert_params(g, State.TERM, fwd=True)

    @pytest.mark.parametrize("state", list(State))
    def test_params(self, adata_large: AnnData, state: State):
        g = _fit_gpcca(adata_large, state)
        _assert_params(g, state, fwd=False)
        _assert_params(g, state, fwd=True)


class TestGPCCASerialization:
    @pytest.mark.parametrize("state", list(State))
    def test_to_adata(self, adata_large: AnnData, state: State):
        g = _fit_gpcca(adata_large, state)
        adata = g.to_adata()
        _assert_adata(adata, state, fwd=False)
        _assert_adata(adata, state, fwd=True)

    @pytest.mark.parametrize("copy", [False, True])
    @pytest.mark.parametrize("keep", ["X", ("obs", "obsm"), ("layers")])
    def test_to_adata_keep(
        self, adata_large: AnnData, keep: Union[str, Sequence[str]], copy: bool
    ):
        g = _fit_gpcca(adata_large, State.MACRO)
        adata = g.to_adata(keep=keep, copy=copy)
        if "X" in keep:
            res = shares_mem(adata.X, g.adata.X)
            assert not res if copy else res
        if "obs" in keep:
            # we don't save macrostates in the underlying object, only during serialization
            key = Key.obs.macrostates(g.backward)
            cols = [c for c in adata.obs.columns if c != key]
            obs = adata.obs[cols]
            np.testing.assert_array_equal(obs.shape, g.adata.obs.shape)
            assert_frame_equal(obs, g.adata.obs[cols])
        if "obsm" in keep:
            for key in g.adata.obsm.keys():
                res = shares_mem(adata.obsm[key], g.adata.obsm[key])
                assert not res if copy else res, key
        if "layers" in keep:
            for key in g.adata.layers.keys():
                res = shares_mem(adata.layers[key], g.adata.layers[key])
                assert not res if copy else res, key

    @pytest.mark.parametrize("copy", [False, True])
    def test_to_adata_copy_finegrained(self, adata_large: AnnData, copy):
        g = _fit_gpcca(adata_large, State.MACRO)
        adata = g.to_adata(keep=["obsp", "varm", "var"], copy=["obsp"])

        assert_frame_equal(adata.var, g.adata.var)
        for key in g.adata.obsp.keys():
            assert not shares_mem(adata.obsp[key], g.adata.obsp[key]), key
        for key in g.adata.varm.keys():
            assert shares_mem(adata.varm[key], g.adata.varm[key]), key

    @pytest.mark.parametrize("state", list(State))
    def test_from_adata(self, adata_large: AnnData, state: State):
        g1 = _fit_gpcca(adata_large, state)
        adata = g1.to_adata()
        g2 = cr.tl.estimators.GPCCA.from_adata(adata, obsp_key="T_fwd")

        _assert_adata(g2.adata, state, fwd=False)
        _assert_adata(g2.adata, state, fwd=True)
        assert_estimators_equal(g1, g2, from_adata=True)

    @pytest.mark.parametrize("state", list(State))
    def test_from_adata_incomplete(self, adata_large: AnnData, state: State):
        if state in (State.SCHUR, State.MACRO):
            pytest.skip("See the reintroduce TODO in GPCCA")
        g_orig = _fit_gpcca(adata_large, State.LIN)
        adata = g_orig.to_adata()

        for attr, key in state.attr_keys:
            obj = getattr(adata, attr)
            del obj[key]

        g = cr.tl.estimators.GPCCA.from_adata(adata, obsp_key="T_fwd")
        _assert_gpcca_attrs(g, state, fwd=True, init=False)
        _assert_gpcca_attrs(g, state.prev, fwd=False)
        _assert_adata(g._shadow_adata, state, fwd=True, init=False)
        _assert_adata(g._shadow_adata, state.prev, fwd=False)

    @pytest.mark.parametrize("bwd", [False, True])
    def test_reconstruct_lineage(self, adata_large: AnnData, bwd: bool):
        g_orig = _fit_gpcca(adata_large, State.LIN, backward=bwd)
        adata = g_orig.to_adata()
        bdata = adata.copy()
        keys = [
            Key.obsm.memberships(Key.obs.macrostates(bwd)),
            Key.obsm.memberships(Key.obs.term_states(bwd)),
            Key.obsm.abs_probs(bwd),
        ]
        attrs = [
            "macrostates_memberships",
            "terminal_states_memberships",
            "absorption_probabilities",
        ]

        for key in keys:
            assert isinstance(adata.obsm[key], Lineage)
            adata.obsm[key] = np.array(adata.obsm[key])
            assert isinstance(adata.obsm[key], np.ndarray)

        g = cr.tl.estimators.GPCCA.from_adata(adata, obsp_key=Key.uns.kernel(bwd))
        for attr, key in zip(attrs, keys):
            obj = getattr(g, attr)
            assert isinstance(obj, Lineage)
            np.testing.assert_allclose(obj.X, bdata.obsm[key].X)
            np.testing.assert_array_equal(obj.names, bdata.obsm[key].names)
            np.testing.assert_array_equal(obj.colors, bdata.obsm[key].colors)

        assert_estimators_equal(g_orig, g, from_adata=True)


class TestGPCCAIO:
    @pytest.mark.parametrize("deep", [False, True])
    def test_copy(
        self, adata_gpcca_fwd: Tuple[AnnData, cr.tl.estimators.GPCCA], deep: bool
    ):
        _, mc1 = adata_gpcca_fwd
        mc2 = mc1.copy(deep=deep)

        assert_estimators_equal(mc1, mc2, copy=True, deep=deep)

    def test_write_ext(
        self, adata_gpcca_fwd: Tuple[AnnData, cr.tl.estimators.GPCCA], tmpdir
    ):
        _, mc = adata_gpcca_fwd

        fname = "foo"
        mc.write(os.path.join(tmpdir, fname), ext="bar")

        assert os.path.isfile(os.path.join(tmpdir, f"foo.bar"))

    def test_write_no_ext(
        self,
        adata_gpcca_fwd: Tuple[AnnData, cr.tl.estimators.GPCCA],
        tmpdir,
    ):
        _, mc = adata_gpcca_fwd
        fname = "foo"
        mc.write(os.path.join(tmpdir, fname), ext=None)

        assert os.path.isfile(os.path.join(tmpdir, f"foo"))

    def test_write_ext_with_dot(
        self,
        adata_gpcca_fwd: Tuple[AnnData, cr.tl.estimators.GPCCA],
        tmpdir,
    ):
        _, mc = adata_gpcca_fwd

        fname = "foo"
        mc.write(os.path.join(tmpdir, fname), ext=".bar")

        assert os.path.isfile(os.path.join(tmpdir, f"foo.bar"))

    def test_read(
        self, adata_gpcca_fwd: Tuple[AnnData, cr.tl.estimators.GPCCA], tmpdir
    ):
        _, mc1 = adata_gpcca_fwd

        mc1.write(os.path.join(tmpdir, "foo"))
        mc2 = cr.tl.estimators.GPCCA.read(os.path.join(tmpdir, "foo.pickle"))

        assert_estimators_equal(mc1, mc2)

    @pytest.mark.parametrize("copy", [False, True])
    def test_write_no_adata(
        self,
        adata_gpcca_fwd: Tuple[AnnData, cr.tl.estimators.GPCCA],
        copy: bool,
        tmpdir,
    ):
        adata, mc1 = adata_gpcca_fwd

        mc1.write(os.path.join(tmpdir, "foo"), write_adata=False)
        mc2 = cr.tl.estimators.GPCCA.read(
            os.path.join(tmpdir, "foo.pickle"), adata=adata, copy=copy
        )

        if copy:
            assert adata is not mc2.adata
        else:
            assert adata is mc2.adata
        assert_estimators_equal(mc1, mc2)

    def test_write_no_adata_read_none_supplied(
        self, adata_gpcca_fwd: Tuple[AnnData, cr.tl.estimators.GPCCA], tmpdir
    ):
        _, mc1 = adata_gpcca_fwd

        mc1.write(os.path.join(tmpdir, "foo"), write_adata=False)
        with pytest.raises(TypeError, match="This object was saved without"):
            _ = cr.tl.estimators.GPCCA.read(
                os.path.join(tmpdir, "foo.pickle"), adata=None
            )

    def test_write_no_adata_read_wrong_length(
        self, adata_gpcca_fwd: Tuple[AnnData, cr.tl.estimators.GPCCA], tmpdir
    ):
        _, mc1 = adata_gpcca_fwd
        adata = AnnData(np.random.normal(size=(len(mc1) + 1, 1)))

        mc1.write(os.path.join(tmpdir, "foo"), write_adata=False)
        with pytest.raises(ValueError, match="Expected `adata` to be of length"):
            _ = cr.tl.estimators.GPCCA.read(
                os.path.join(tmpdir, "foo.pickle"), adata=adata
            )



from typing import Any, Union, Optional, Sequence
from typing_extensions import Literal

import cellrank.logging as logg
from cellrank._key import Key
from cellrank.ul._docs import d
from cellrank.pl._utils import AnnData
from cellrank.tl.estimators.terminal_states._cflare import CFLARE


@d.dedent
def lineages(
    adata: AnnData,
    lineages: Optional[Union[str, Sequence[str]]] = None,
    backward: bool = False,
    color: Optional[str] = None,
    mode: Literal["embedding", "time"] = "embedding",
    time_key: str = "latent_time",
    **kwargs: Any,
) -> None:
    """
    Plot lineages that were uncovered using :func:`cellrank.tl.lineages`.

    For each lineage, we show all cells in an embedding (default is UMAP) and color them by their probability of
    belonging to this lineage. For cells that are already committed, this probability will be one for  their respective
    lineage and zero otherwise. For naive cells, these probabilities will be more balanced, reflecting
    the fact that naive cells have the potential to develop towards multiple endpoints.

    Parameters
    ----------
    %(adata)s
    lineages
        Plot only these lineages. If `None`, plot all lineages.
    %(backward)s
    color
        If given, plot cluster annotations left of the lineage probabilities.
    %(time_mode)s
    time_key
        Key in ``adata.obs`` where the pseudotime is stored.
    %(basis)s
    kwargs
        Keyword arguments for :meth:`cellrank.tl.estimators.GPCCA.plot_absorption_probabilities`.

    Returns
    -------
    %(just_plots)s
    """

    # use CFLARE (no macrostates required)
    mc = CFLARE.from_adata(adata, obsp_key=Key.uns.kernel(backward))
    if mc.absorption_probabilities is None:
        raise RuntimeError(
            f"Compute absorption probabilities first as `cellrank.tl.lineages(..., backward={backward})`."
        )

    # plot using the MC object
    color = kwargs.pop("cluster_key", color)
    mc.plot_absorption_probabilities(
        states=lineages,
        color=color,
        mode=mode,
        time_key=time_key,
        **kwargs,
    )


@d.dedent
def lineage_drivers(
    adata: AnnData,
    lineage: str,
    backward: bool = False,
    n_genes: int = 8,
    ncols: Optional[int] = None,
    use_raw: bool = False,
    title_fmt: str = "{gene} qval={qval:.4e}",
    **kwargs: Any,
) -> None:
    """
    Plot lineage drivers that were uncovered using :func:`cellrank.tl.lineage_drivers`.

    Parameters
    ----------
    %(adata)s
    %(backward)s
    %(plot_lineage_drivers.parameters)s

    Returns
    -------
    %(just_plots)s
    """

    mc = CFLARE.from_adata(adata, obsp_key=Key.uns.kernel(backward))

    if use_raw and adata.raw is None:
        logg.warning("No raw attribute set. Using `use_raw=False`")
        use_raw = False

    key = Key.varm.lineage_drivers(backward)
    haystack = (adata.raw if use_raw else adata).varm

    if key not in haystack:
        raise RuntimeError(
            "Unable to find lineage drivers in "
            f"`{'adata.raw.varm' if use_raw else 'adata.varm'}[{key!r}]`. "
            f"Compute lineage drivers first as `cellrank.tl.lineage_drivers(lineages={lineage!r}, "
            f"use_raw={use_raw}, backward={backward}).`"
        )

    mc._lineage_drivers = haystack[key]
    mc.plot_lineage_drivers(
        lineage,
        n_genes=n_genes,
        use_raw=use_raw,
        ncols=ncols,
        title_fmt=title_fmt,
        **kwargs,
    )
